#!/usr/bin/env python
"""
Minor version changes:
 - cleaned up
Input:
- genbank file
- dictionary
- specific features : 16S, 5S
- upstream, downstream widths

Output:
-dir containing DNA fastas in their

"""
# import re
import os
# import csv
import subprocess
import datetime
import time
import argparse
import sys
import math
import re
import shutil
import itertools
import multiprocessing

import numpy as np
import matplotlib as mpl
mpl.use('Agg')
# from matplotlib.backends.backend_pdf import PdfPages
import matplotlib.pyplot as plt
import matplotlib.patches as patches
import pandas as pd

from Bio import SeqIO
from Bio.SeqRecord import SeqRecord
from Bio.Alphabet import IUPAC
from Bio.Blast.Applications import NcbiblastnCommandline
from Bio.Seq import Seq
from collections import defaultdict  # for calculating kmer frequency
from itertools import product  # for getting all possible kmers
from heatmapcluster import heatmapcluster

#from pyutilsnrw import utils3_5
from pyutilsnrw.utils3_5 import set_up_logging, check_installed_tools,\
    combine_contigs, check_version_from_cmd


class LociCluster(object):
    """ organizes the clustering process instead of dealing with nested lists
    This holds the whole cluster of one to several individual loci
    """
    newid = itertools.count()

    def __init__(self, sequence_id, loci_list, padding=None,
                 global_start_coord=None, global_end_coord=None,
                 seq_record=None, feat_of_interest=None, mappings=None,
                 extractedSeqRecord=None, cluster_dir_name=None,
                 circular=False, output_root=None, final_contigs_path=None,
                 continue_iterating=True, keep_contigs=True):
        # int: unique identifier for cluster
        self.index = next(LociCluster.newid)
        # self.index = index
        # str: sequence name, usually looks like 'NC_17777373.1' or similar
        self.sequence_id = sequence_id
        # list: hold locus objects for each item in cluster
        self.loci_list = loci_list  # this holds the Locus objects
        # int: bounds ____[___.....rRNA....rRNA..rRNA....__]_________
        self.global_start_coord = global_start_coord
        self.global_end_coord = global_end_coord
        # int: how much to pad sequences y if treating as circular
        self.padding = padding
        # str: feature for filtering: rRNA, cDNA, exon, etc
        self.feat_of_interest = feat_of_interest
        # Bool: treat seqs as circular by padding the ends
        self.circular = circular
        # path: where your cluster-specific output goes
        self.cluster_dir_name = cluster_dir_name  # named dynamically
        # path: where the overall output goes
        self.output_root = output_root
        # list: lociMapping objects that hold mappinging paths
        self.mappings = mappings
        # SeqRecord: holds SeqIO Seqrecord for sequence_id
        self.seq_record = seq_record
        # SeqRecord: holds SeqIO Seqrecord for seq extracted from global coords
        self.extractedSeqRecord = extractedSeqRecord
        # path: for best contig after riboseed2 iterations
        self.keep_contigs = keep_contigs  # by default, include all
        self.continue_iterating = continue_iterating  # by default, keep going

        self.final_contig_path = final_contigs_path
        self.name_mapping_dir()

    def name_mapping_dir(self):
        self.cluster_dir_name = str("{0}_cluster_{1}").format(
            self.sequence_id, self.index)


class Locus(object):
    """ this holds the info for each individual Locus"
    """
    def __init__(self, index, sequence_id, locus_tag, strand=None,
                 start_coord=None, end_coord=None, rel_start_coord=None,
                 rel_end_coord=None, product=None):
        # int: unique identifier for cluster
        self.index = index
        # str: sequence name, usually looks like 'NC_17777373.1' or similar
        self.sequence_id = sequence_id  # is this needed? I dont think so as long
        # str: unique identifier from \locus_tag= of gb file
        self.locus_tag = locus_tag
        # int: 1 is +strand, -1 is -strand
        self.strand = strand
        # int:
        self.start_coord = start_coord
        self.end_coord = end_coord
        # str: from \product= of gb file
        self.product = product


def get_args():  # pragma: no cover
    """get the arguments as a main parser with subparsers
    for named required arguments and optional arguments
    """
    parser = argparse.ArgumentParser(description="Use to extract regions " +
                                     "of interest based on supplied Locus " +
                                     " tags.", add_help=False)
    parser.add_argument("genbank_genome", help="Genbank file (WITH SEQUENCE)")
    parser.add_argument("clustered_loci", help="output from riboSelect")

    requiredNamed = parser.add_argument_group('required named arguments')
    requiredNamed.add_argument("-o", "--output",
                               help="output directory; default: %(default)s",
                               default=os.getcwd(),
                               type=str, dest="output")

    # had to make this faux "optional" parse so that the named required ones
    # above get listed first
    optional = parser.add_argument_group('optional arguments')
    # parser.add_argument("-f", "--feature", help="Feature, such as CDS,tRNA, " +
    #                     "rRNA; default: %(default)s",
    #                     default='rRNA', dest="feature",
    #                     action="store", type=str)
    # optional.add_argument("-w", "--within_feature_length",
    #                       help="bp's to include within the region; " +
    #                       "default: %(default)s",
    #                       default=0, dest="within", action="store", type=int)
    optional.add_argument("-n", "--name",
                          help="rename the contigs with this prefix" +
                          # "default: %(default)s",
                          "default: date (YYYYMMDD)",
                          default=None, dest="name",
                          action="store", type=str)
    optional.add_argument("-l", "--flanking_length",
                          help="length of flanking regions, in bp; " +
                          "default: %(default)s",
                          default=1000, type=int, dest="flanking")
    optional.add_argument("--msa_kmers",
                          help="calculate kmer similarity based on aligned " +
                          "sequences instead of raw sequences;" +
                          "default: %(default)s",
                          default=False, action="store_true", dest="msa_kmers")
    optional.add_argument("-c", "--circular",
                          help="if the genome is known to be circular, and " +
                          "an region of interest (including flanking bits) " +
                          "extends past chromosome end, this extends the " +
                          "seqence past chromosome origin forward by 5kb; " +
                          "default: %(default)s",
                          default=False, dest="circular", action="store_true")
    optional.add_argument("-p", "--padding", dest='padding', action="store",
                          default=5000, type=int,
                          help="if treating as circular, this controls the " +
                          "length of sequence added to the 5' and 3' ends " +
                          "to allow for selecting regions that cross the " +
                          "chromosom's origin; default: %(default)s")
    optional.add_argument("-v", "--verbosity",
                          dest='verbosity', action="store",
                          default=2, type=int,
                          help="1 = debug(), 2 = info(), 3 = warning(), " +
                          "4 = error() and 5 = critical(); " +
                          "default: %(default)s")
    optional.add_argument("--clobber",
                          help="overwrite previous output files" +
                          "default: %(default)s", action='store_true',
                          default=False, dest="clobber")
    optional.add_argument("--no_revcomp",
                          help="default returns reverse complimented seq " +
                          "if majority of regions on reverse strand. if  " +
                          "--no_revcomp, this is overwridden" +
                          "default: %(default)s",
                          action='store_true',
                          default=False, dest="no_revcomp")
    optional.add_argument("--skip_check",
                          help="Dont bother calculating Shannon Entropy; " +
                          "default: %(default)s",
                          action='store_true',
                          default=False, dest="skip_check")
    optional.add_argument("--msa_tool", dest="msa_tool",
                          choices=["mafft", "prank"],
                          action="store", default="mafft",
                          help="Path to PRANK executable; " +
                          "default: %(default)s")
    optional.add_argument("--prank_exe", dest="prank_exe",
                          action="store", default="prank",
                          help="Path to PRANK executable; " +
                          "default: %(default)s")
    optional.add_argument("--mafft_exe", dest="mafft_exe",
                          action="store", default="mafft",
                          help="Path to MAFFT executable; " +
                          "default: %(default)s")
    optional.add_argument("--barrnap_exe", dest="barrnap_exe",
                          action="store", default="barrnap",
                          help="Path to barrnap executable; " +
                          "default: %(default)s")
    optional.add_argument("--makeblastdb_exe", dest="makeblastdb_exe",
                          action="store", default="makeblastdb",
                          help="Path to makeblastdb executable; " +
                          "default: %(default)s")
    optional.add_argument("--kingdom", dest="kingdom",
                          action="store", default="bac",
                          choices=["mito", "euk", "arc", "bac"],
                          help="kingdom for barrnap; " +
                          "default: %(default)s")
    # had to make this explicitly to call it a faux optional arg
    optional.add_argument("-h", "--help",
                          action="help", default=argparse.SUPPRESS,
                          help="Displays this help message")
    args = parser.parse_args()
    return args


def get_genbank_rec_from_multigb(recordID, genbank_records):
    """ given a record ID and and list of genbank records, return sequence of
    genbank record that has all the loci.
    If on different sequences, return error
    """
    for record in genbank_records:
        if recordID == record.id:
            return record
        else:
            pass
    # if none found, raise error
    raise ValueError("no record found matching record id %s!" % recordID)


def parse_clustered_loci_file(filepath, gb_filepath, output_root,
                              padding, circular, logger=None):
    """Given a file from riboSelect or manually created (see specs in README)
    this parses the clusters and returns a list where [0] is sequence name
    and [1] is a list of loci in that cluster
    As of 20161028, this returns a list of LociCluster objects!
    """
    if logger is None:
        raise ValueError("logging must be used!")
    if not (os.path.isfile(filepath) and os.path.getsize(filepath) > 0):
        raise ValueError("Cluster File not found!")
    clusters = []
    # this covers common case where user submits genbank and cluster file
    # in the wrong order.
    if filepath.endswith(("gb", "genbank", "gbk")):
        raise FileNotFoundError("Hmm, this cluster file looks like genbank; " +
                                "it ends in {0}".format(os.path.splitext(
                                    filepath)[1]))
    try:
        with open(filepath, "r") as f:
            file_contents = list(f)
    except Exception as e:
        logger.error("Cluster file could not be parsed!")
        raise e
    feature = None
    for line in file_contents:
        try:
            if line.startswith("#$ FEATURE"):
                try:
                    feature = line.split("FEATURE")[1].strip()
                except:
                    raise ValueError("Cannot extract FEATURE from '%s'" % line)
                continue
            elif line.startswith("#") or line.strip() == '':
                continue
            seqname = line.strip("\n").split(" ")[0]
            lt_list = [x for x in
                       line.strip("\n").split(" ")[1].split(":")]
        except Exception as e:
            logger.error("error parsing line: %s" % line)
            raise e
        # make and append the locus objects
        loci_list = []
        for i, loc in enumerate(lt_list):
            loci_list.append(Locus(index=i,
                                   locus_tag=loc,
                                   sequence_id=seqname))
        # make and append LociCluster objects
        clusters.append(LociCluster(mappings=[],
                                    output_root=output_root,
                                    sequence_id=seqname,
                                    loci_list=loci_list,
                                    padding=padding,
                                    circular=circular))
        # cluster_index = cluster_index + 1

    ### check feature; if still none or starts with #$ (ie, no split)
    if feature is None:
        raise ValueError("no feature extracted from coords file! This " +
                         "has been made mandatory 20161108")
    ###
    if len(clusters) == 0:
        raise ValueError("No Clusters Found!!")
    # match up seqrecords
    # with open(gb_filepath) as fh:
    #     gb_records = list(SeqIO.parse(fh, 'genbank'))
    gb_records = SeqIO.index(gb_filepath, 'genbank')
    for clu in clusters:
        clu.feat_of_interest = feature
        clu.seq_record = gb_records[clu.sequence_id]
        # clu.seq_record = get_genbank_rec_from_multigb(
        #     recordID=clu.sequence_id,
        #     genbank_records=gb_records)
    gb_records.close()
    return clusters


def extract_coords_from_locus(cluster, feature="rRNA",
                              logger=None, verbose=False):
    """given a LociCluster object, ammend values
    20161028 returns a LociCluster
    """
    if logger is None:
        raise ValueError("logging must be used!")
    loc_number = 0  # index for hits
    locus_tags = [x.locus_tag for x in cluster.loci_list]
    for feat in cluster.seq_record.features:
        if not feat.type in feature:
            continue
        if verbose:
            logger.debug("found {0} in the following feature : \n{1}".format(
                feature, feat))
        try:
            locus_tag = feat.qualifiers.get("locus_tag")[0]
        except:
            logger.error(str("found a feature ({0}), but there is no" +
                             "locus tag associated with it! Try formatting " +
                             "it by running scanScaffolds.sh").format(feat))
            raise ValueError

        # quick way of checking without using whole object
        if locus_tag in locus_tags:
            # make this_locus point to locus we are adding info to
            this_locus = next((x for x in cluster.loci_list if
                               x.locus_tag == locus_tag), None)
            #  SeqIO makes coords 0-based; the +1 below undoes that
            this_locus.start_coord = feat.location.start.position + 1
            this_locus.end_coord = feat.location.end.position
            this_locus.strand = feat.strand
            this_locus.product = feat.qualifiers.get("product")
            logger.debug("Added attributes for %s", this_locus.locus_tag)
            # logger.debug(str(this_locus.__dict__))
            logger.debug(
                "locus_tag: %s; coords: [%i-%i]; sequence_id: %s; " +
                "product: %s", this_locus.locus_tag,
                this_locus.start_coord, this_locus.end_coord,
                this_locus.sequence_id, this_locus.product[0])
            loc_number = loc_number + 1
        else:
            pass
            # logger.error("skipping %s", locus_tag)
    if not loc_number > 0:
        logger.error("no hits found in any record with feature %s! Double " +
                     "check your genbank file", feature)
        raise ValueError
    # logger.debug("adding extracted sequence")
    # cluster.extractedSeqRecord = cluster.seq_record.seq[cluster.
    # return cluster


def pad_genbank_sequence(cluster, logger=None, verbose=False):
    """coords in coords list should be the 1st list item, with the index
    being 0th. Given a genbank record and a coord_list. this returns a seq
    padded on both ends by --padding bp, and returns a coord_list with coords
    adjusted accordingly.  Used to capture regions across origin.
    # as of 20161028, cluster object, not coord_list, is used
    """
    # take care of the coordinates
    if logger:
        logger.info(str("adjusting coordinates by {0} to account for " +
                        "padding").format(cluster.padding))
    for loc in cluster.loci_list:
        if logger:
            logger.debug("pre-padded")
            logger.debug(str(loc.__dict__))
        start, end = loc.start_coord, loc.end_coord
        loc.start_coord, loc.end_coord = [start + cluster.padding,
                                          end + cluster.padding]
        if logger:
            logger.debug("post-padded")
            logger.debug(str(loc.__dict__))
    #TODO: check for interference with other clusters

    # take care of the sequence
    old_seq = cluster.seq_record.seq
    if cluster.padding > len(old_seq):
        raise ValueError("padding cannot be greater than length of sequence")
    new_seq = str(old_seq[-cluster.padding:]
                  + old_seq
                  + old_seq[0: cluster.padding])
    assert len(new_seq) == (len(old_seq) + (2 * cluster.padding)), \
        "Error within function! new seq should be len of " + \
        "seq plus 2x padding"
    cluster.seq_record = SeqRecord(Seq(new_seq))
    return cluster


def stitch_together_target_regions(cluster,
                                   flanking=500,
                                   logger=None, circular=False,
                                   revcomp=False):
    """
    given a single LociCluster object containing Locus objects (usually)
    of length 3 (16,5,and 23 rRNAs),
    return the object with ammended sequence info
    revamped 20161004
    """
    if logger is None:
        raise ValueError("Must have logger for this function")

    # TODO : make this safer. coord list is constructed sequentially but this
    # is a backup. Throws sort of a cryptic error. but as I said, its a backup
    if sorted([x.start_coord for x in cluster.loci_list]) != \
       [x.start_coord for x in cluster.loci_list]:
        logger.warning("Coords are not in increasing order; " +
                       "you've been warned")
    start_list = sorted([x.start_coord for x in cluster.loci_list])
    logger.debug("Start_list: {0}".format(start_list))

    logger.debug("stitching together the following coords:")
    for i in cluster.loci_list:
        logger.debug(str(i.__dict__))
    #  This works as long as coords are never in reverse order
    cluster.global_start_coord = min([x.start_coord for
                                      x in cluster.loci_list]) - flanking
    # if start is negative, just use 1, the beginning of the sequence
    if cluster.global_start_coord < 1:
        logger.warning("Caution! Cannot retrieve full flanking region, as " +
                       "the 5' flanking region extends past start of " +
                       "sequence. If this is a problem, try using a smaller " +
                       "--flanking region, and/or if  appropriate, run with " +
                       "--circular.")
        cluster.global_start_coord = 1
    cluster.global_end_coord = max([x.end_coord for
                                    x in cluster.loci_list]) + flanking
    if cluster.global_end_coord > len(cluster.seq_record):
        logger.warning("Caution! Cannot retrieve full flanking region, as " +
                       "the 5' flanking region extends past start of " +
                       "sequence. If this is a problem, try using a smaller " +
                       "--flanking region, and/or if  appropriate, run with " +
                       "--circular.")
        cluster.global_end_coord = len(cluster.seq_record)

    logger.debug("global start and end: %s %s", cluster.global_start_coord,
                 cluster.global_end_coord)
    #  the minus one makes things go from 1 based to zero based
    seq_with_ns = str(cluster.seq_record.seq[cluster.global_start_coord - 1:
                                             cluster.global_end_coord])
    seq_len = len(seq_with_ns[:])
    logger.info(seq_len)
    # again, plus 1 corrects for 0 index.
    # len("AAAA") = 4 vs AAAA[-1] - AAAA[0] = 3
    logger.info("\nexp length %i \nact length %i",
                cluster.global_end_coord - cluster.global_start_coord + 1,
                seq_len)

    ## Change coords in ID When using padding
    if not circular:
        seq_id = str(cluster.sequence_id + "_" + str(cluster.global_start_coord) +
                     ".." + str(cluster.global_end_coord))
    else:  # correct for padding
        seq_id = str(cluster.sequence_id + "_" + str(cluster.global_start_coord -
                                                     cluster.padding) +
                     ".." + str(cluster.global_end_coord - cluster.padding))

    # if most are on - strand, return sequence reverse complement
    strands = [x.strand for x in cluster.loci_list]
    if revcomp and \
       (sum([x == -1 for x in strands]) > sum([x == 1 for x in strands])):
        logger.info("returning the reverse compliment of the sequence")
        cluster.extractedSeqRecord = SeqRecord(
            Seq(seq_with_ns,
                IUPAC.IUPACAmbiguousDNA()).reverse_complement(),
            id=str(seq_id + "_RC"))
    else:
        cluster.extractedSeqRecord = SeqRecord(
            Seq(seq_with_ns,
                IUPAC.IUPACAmbiguousDNA()),
            id=seq_id)
    ### last minuete check
    for property, value in vars(cluster).items():
        if value is None:
            logger.debug("%s has a value of None!", property)
    return cluster


def prepare_prank_cmd(outdir, combined_fastas, prank_exe,
                      add_args="", outfile_name="best_MSA",
                      logger=None):
    """returns command line for constructing MSA with
    PRANK and the path to results file
    """
    if logger is None:
        raise ValueError("Must use logger")
    if not os.path.exists(outdir):
        raise FileNotFoundError("output directory not found!")
    prank_cmd = "{0} {1} -d={2} -o={3}".format(
        prank_exe, add_args, combined_fastas,
        os.path.join(outdir, outfile_name))
    logger.debug("PRANK command: \n %s", prank_cmd)
    return (prank_cmd, os.path.join(outdir, str(outfile_name + ".fasta")))


def prepare_mafft_cmd(outdir, combined_fastas, mafft_exe,
                      add_args="", outfile_name="best_MSA",
                      logger=None):
    """returns command line for constructing MSA with
    mafft and the path to results file
    """
    if logger is None:
        raise ValueError("Must use logger")
    if not os.path.exists(outdir):
        raise FileNotFoundError("output directory not found!")
    mafft_cmd = "{0} {1} {2} > {3}".format(
        mafft_exe, add_args, combined_fastas,
        os.path.join(outdir, outfile_name))
    logger.debug("MAFFT command: \n %s", mafft_cmd)
    return (mafft_cmd, os.path.join(outdir, outfile_name))


def calc_Shannon_entropy(matrix):
    """ $j$ has entropy $H(j)$ such that
    $H(j) = -sum_{i=(A,C,T,G)} p_i(j) log p_i(j)$
    """
    entropies = []
    for instance in matrix:
        unique = set(instance)
        proportions = {}
        for i in unique:
            proportions[i] = sum([x == i for x in instance]) / len(instance)
        entropy = -sum([prob * (math.log(prob, math.e)) for
                        prob in proportions.values()])
        entropies.append(entropy)
    return entropies


def calc_entropy_msa(msa_path):
    """givn a path to an MSA in FASTA format, this gets the
    $j$ has entropy $H(j)$ such that
    $H(j) = -sum_{i=(A,C,T,G)} p_i(j) log p_i(j)$
    return list
    """
    batch_size = 1000  # read seequences in chunks this long
    lengths = []
    with open(msa_path) as fh:
        msa_seqs = list(SeqIO.parse(fh, 'fasta'))
    seq_names = []
    for rec in msa_seqs:
        lengths.append(len(rec))
        seq_names.append(rec.id)
    if not all([i == lengths[0] for i in lengths]):
        raise ValueError("Sequences must all be the same length!")
    entropies = []
    tseq = []
    # save memory by reading in chunks
    for batch in range(0, (math.ceil(lengths[0] / batch_size))):
        # get all sequences into an array
        seq_array = []
        for nseq, record in enumerate(msa_seqs):
            seq_array.append(
                [x for x in record.seq[(batch * batch_size):
                                       ((batch + 1) * batch_size)]])
        # transpose
        tseq_array = list(map(list, zip(*seq_array)))
        tseq.extend(tseq_array)
        entropies.extend(calc_Shannon_entropy(tseq_array))
    # check length of sequence is the same as length of the entropies
    assert len(entropies) == lengths[0]
    return (entropies, seq_names, tseq)


def annotate_msa_conensus(tseq_array, seq_file, barrnap_exe,
                          kingdom="bact",
                          pattern='product=(.+?)$',
                          countdashcov=True,   # include -'s in coverage
                          collapseNs=False,  # include -'s in consensus
                          excludedash=False,
                          logger=None):
    """ returns annotations (as gfflist),the consensus sequence as a
    list[base, cov], and named coords  as a list
    TODO: The 'next_best' thing fails is an N is most frequent. Defaults to a T
    """
    if logger is None:
        raise ValueError("Must use logging")
    if excludedash:
        logger.warning("CAUTION: excludedash selected. There is a known " +
                       "bug in the 'next_best' thing fails if an " +
                       "N is most frequent. Defaults to a T")
    consensus = []
    nseqs = len(tseq_array[0])
    logger.info("calc coverage for each of the %i positions", len(tseq_array))
    for position in tseq_array:
        if all([x == position[0] for x in position]):
            if position[0] == '-':
                if collapseNs:
                    continue
                elif not countdashcov:
                    consensus.append([position[0], 0])
            else:
                consensus.append([position[0], nseqs])
        else:
            max_count = 0  # starting count
            nextbest_count = 0
            best_nuc = None
            nextbest_nuc = None
            for nuc in set(position):
                count = sum([nuc == z for z in position])
                # if max count, swap with max to nextbest and update max
                if count > max_count:
                    nextbest_count = max_count
                    max_count = count  # update count if better
                    nextbest_nuc = best_nuc
                    best_nuc = nuc
                else:
                    pass
            # contol whether gaps are allowed in consensus
            if (
                    all([x != '-' for x in position]) and
                    best_nuc == '-' and
                    excludedash):
                # if we dont want n's, choose nextbest
                if nextbest_nuc is None:
                    nextbest_nuc = 't'  # I hate myself for this
                consensus.append([nextbest_nuc, nextbest_count])
            elif best_nuc == '-' and not countdashcov:
                consensus.append([best_nuc, 0])
            else:
                consensus.append([best_nuc, max_count])
    # if any are '-', replace with n's for barrnap
    seq = str(''.join([x[0] for x in consensus])).replace('-', 'n')

    # annotate seq
    with open(seq_file, 'w') as output:
        SeqIO.write(SeqRecord(
            Seq(seq, IUPAC.IUPACAmbiguousDNA())), output, "fasta"),
    barrnap_cmd = "{0} --kingdom {1} {2}".format(barrnap_exe,
                                                 kingdom, seq_file)
    barrnap_gff = subprocess.run(barrnap_cmd,
                                 shell=sys.platform != "win32",
                                 stdout=subprocess.PIPE,
                                 stderr=subprocess.PIPE,
                                 check=True)
    results_list = [x.split('\t') for x in
                    barrnap_gff.stdout.decode("utf-8").split("\n")]

    ###  make [name, [start_coord, end_coord]] list
    named_coords = []
    for i in results_list:
        if i[0].startswith("#") or not len(i) == 9:
            logger.debug("skipping gff line: %s", i)
            continue
        m = re.search(pattern, i[8])
        if m:
            found = m.group(1)
        named_coords.append([found, [int(i[3]), int(i[4])]])
    if len(named_coords) == 0:
        raise ValueError(str("Error extracting coords from barrnap gff line" +
                             " %s with pattern %s!" % (str(i), pattern)))
    return (results_list, consensus, named_coords)


def plot_scatter_with_anno(data,
                           consensus_cov,
                           anno_list,
                           names=["Position", "Entropy"],
                           title="Shannon Entropy by Position",
                           output_prefix="entropy_plot.png"):
    """Given annotation coords [feature, [start, end]],
    consensus cov list ['base', coverage_depth],
    entropy values (list) and consensus sequence
    (same length for consensus_cov and data, no funny business),
    plot out the entropies for each position,
    plot the annotations, and return 0
    """
    if len(consensus_cov) != len(data):
        raise ValueError("data and consensus different lengths!")
    df = pd.DataFrame({names[0]: range(1, len(data) + 1),
                       names[1]: data})  # columns=names)
    df_con = pd.DataFrame(consensus_cov, columns=["base",
                                                  "depth"])
    cov_max_depth = max(df_con['depth'])
    fig, (ax1, ax2) = plt.subplots(2, 1, sharex=True,
                                   gridspec_kw={'height_ratios': [4, 1]})
    colors = ['#FF8306', '#FFFB07', '#04FF08', '#06B9FF', '#6505FF', '#FF012F',
              '#FF8306', '#FFFB07', '#04FF08', '#06B9FF', '#6505FF', '#FF012F']
    ax1.set_title(title, y=1.08)
    xmin, xmax = 0, len(data)
    ymin, ymax = -0.1, (max(data) * 1.2)
    ax1.set_xlim([xmin, xmax])
    ax1.set_ylim([ymin, ymax])
    yjust = -.1
    for index, anno in enumerate(anno_list):
        rect1 = patches.Rectangle(
            (anno[1][0],  # starting x
             ymin),  # starting y
            anno[1][1] - anno[1][0],  # rel x end
            ymax - ymin,  # rel y end
            facecolor=mpl.colors.ColorConverter().to_rgba(
                colors[index], alpha=0.2),
            edgecolor=mpl.colors.ColorConverter().to_rgba(
                colors[index], alpha=0.2))
        rect2 = patches.Rectangle(
            (anno[1][0],  # starting x
             1),  # starting y
            anno[1][1] - anno[1][0],  # rel x end
            cov_max_depth,  # dont -1 beacuse start at 1
            facecolor=mpl.colors.ColorConverter().to_rgba(
                colors[index], alpha=0.2),
            edgecolor=mpl.colors.ColorConverter().to_rgba(
                colors[index], alpha=0.2))
        ax1.add_patch(rect1)
        ax2.add_patch(rect2)
        ax1.text((anno[1][0] + anno[1][1]) / 2,    # x location
                 ymax - 0.48 - yjust,                      # y location
                 anno[0][0:20],                          # text first 20 char
                 ha='center', color='red', weight='bold', fontsize=10)
        yjust = yjust * - 1
    ax1.scatter(x=df["Position"], y=df["Entropy"],
                marker='o', color='black', s=2)
    # add smoothing for kicks
    df["fit"] = savitzky_golay(df["Entropy"].values, 351, 3)  # window size 51, polynomial order 3
    ax1.scatter(x=df["Position"], y=df["fit"], color='red', s=1)
    #
    ax1.set_ylabel('Shannon Entropy')
    ax1.get_yaxis().set_label_coords(-.05, 0.5)
    ax2.set_xlim([xmin, xmax])
    ax2.invert_yaxis()
    ax2.set_ylabel('Consensus Coverage')
    ax2.set_xlabel('Position (bp)')
    ax2.get_yaxis().set_label_coords(-.05, 0.5)
    # ax2.set_ylim([1, cov_max_depth + 1]) #, 1])
    ax2.bar(df_con.index, df_con["depth"],
            width=1, color='darkgrey', linewidth=0, edgecolor='darkgrey')
    # ax2.step(df_con.index, df_con["depth"],
    #          where='mid', color='darkgrey')
    for ax in [ax1, ax2]:
        ax.spines['right'].set_visible(False)
    # Only show ticks on the left and bottom spines
    ax1.spines['top'].set_visible(False)
    ax2.spines['bottom'].set_visible(False)
    ax.yaxis.set_ticks_position('left')
    ax2.xaxis.set_ticks_position('bottom')
    ax1.xaxis.set_ticks_position('top')
    ax1.tick_params(axis='y', colors='dimgrey')
    ax2.tick_params(axis='y', colors='dimgrey')
    ax1.tick_params(axis='x', colors='dimgrey')
    ax2.tick_params(axis='x', colors='dimgrey')
    ax1.yaxis.label.set_color('black')
    ax2.yaxis.label.set_color('black')
    ax1.xaxis.label.set_color('black')
    ax2.xaxis.label.set_color('black')
    plt.tight_layout()
    fig.subplots_adjust(hspace=0)
    fig.set_size_inches(12, 7.5)
    fig.savefig(str(output_prefix + '.png'), dpi=(200))
    fig.savefig(str(output_prefix + '.pdf'), dpi=(200))
    return 0


def get_all_kmers(alph="", length=3):
    """Given an alphabet of charactars, return a list of all permuations
    not actually used, I dont think I'll need this
    """
    mers = [''.join(p) for p in product(alph, repeat=length)]
    return mers


def profile_kmer_occurances(rec_list, k, logger=None):
    """ given a list of seq records, an alphabet, and a value k,
    retrun counts dict of kmer occurances and list of seq names
    """
    counts = defaultdict(list)
    names_list = []
    # part 1: get all kmers from seqs so that we can have equal length lisst
    all_mers = []
    for rec in rec_list:
        all_mers.extend([str(rec.seq).lower()[x: x + k] for
                         x in range(0, len(rec.seq) - k + 1)])
    # unique_mers = list(set().union(all_mers))
    unique_mers = set(all_mers)
    for i in unique_mers:
        counts[i] = []  # initialixe counts dictionary with ker keys
    # part two: count 'em
    for n, rec in enumerate(rec_list):
        logger.info("counting kmer occurances for %s", rec.id)
        names_list.append(rec.id)
        logger.debug("converting to lower")
        string = str(rec.seq).lower()
        logger.debug("getting %imers from seq", k)
        string_mers = [string[x: x + k] for x in range(0, len(string) - k + 1)]
        logger.debug("counting mer occurances")
        # Add counts for those present
        for value in set(string_mers):
            counts[value].append(sum([value == mer for mer in string_mers]))
        # filling in where not found
        for missing in unique_mers - set(string_mers):
            counts[missing].append(0)
    return counts, names_list


def plot_pairwise_least_squares(counts, names_list, output_prefix):
    """given a  list of sequence names, a prefix for plot file names,
    and a list of counts from plot_kmer_occurances,
    retruns a pandas df of least squares after plotting heatmaps
    """
    res_list = []
    counts_list = []
    for v in counts.values():
        counts_list.append(v)
    # this gives each an index and gets all the pairs
    all_pairs = [[index, value] for index, value in
                 enumerate(product(range(0, len(names_list)), repeat=2))]
    for i in all_pairs:
        this_pairs_diffs = []
        for row in counts_list:
            this_pairs_diffs.append((row[i[1][0]] - row[i[1][1]]) ** 2)
        res_list.append([names_list[i[1][0]],
                         names_list[i[1][1]],
                         sum(this_pairs_diffs)])
    lsdf_wNA = pd.DataFrame(res_list, columns=["locus_1", "locus_2", "sls"])
    wlsdf = lsdf_wNA.pivot(index='locus_1', columns='locus_2', values='sls')
    fig, ax = plt.subplots(1, 1)
    lsdf = lsdf_wNA.fillna(value=0)
    heatmap = ax.pcolormesh(wlsdf,
                            cmap='Greens')
    # put the major ticks at the middle of each cell
    ax.set_yticks(np.arange(wlsdf.shape[0]) + 0.5, minor=False)
    ax.set_xticks(np.arange(wlsdf.shape[1]) + 0.5, minor=False)
    # # want a more natural, table-like display
    ax.invert_yaxis()
    ax.xaxis.tick_top()
    plt.xticks(rotation=90)
    # # Set the labels
    ax.set_xticklabels(wlsdf.columns.values, minor=False)
    ax.set_yticklabels(wlsdf.index, minor=False)
    fig.colorbar(heatmap)  # add colorbar key
    plt.tight_layout()  # pad=0, w_pad=5, h_pad=.0)
    fig.set_size_inches(8, 8)
    fig.savefig(str(output_prefix + "heatmap.png"), dpi=(200))
    fig.savefig(str(output_prefix + "heatmap.pdf"), dpi=(200))
    ####  plot clustered heatmap
    plt.close('all')
    plt.figure(1, figsize=(6, 6))
    h = heatmapcluster(wlsdf.as_matrix(), row_labels=wlsdf.index,
                       col_labels=wlsdf.columns.values,
                       num_row_clusters=2, num_col_clusters=2,
                       label_fontsize=6,
                       xlabel_rotation=-75,
                       cmap=plt.cm.coolwarm,
                       show_colorbar=True,
                       top_dendrogram=True)
    # plt.tight_layout()  # pad=0, w_pad=5, h_pad=.0)
    # fig2.set_size_inches(16, 16)
    plt.savefig(str(output_prefix + "clustered_heatmap.png"), dpi=(200))
    plt.savefig(str(output_prefix + "clustered_heatmap.pdf"), dpi=(200))
    return lsdf_wNA


def make_msa(msa_tool, unaligned_seqs, prank_exe, mafft_exe,
             args, outdir, logger=None):
    """returns msa cmd and results path
    """
    if logger is None:
        raise ValueError("Must use logger")
    if msa_tool == "prank":
        if check_installed_tools(executable=prank_exe,
                                 hard=False,
                                 logger=logger):
            msa_cmd, results_path = prepare_prank_cmd(
                outdir=outdir,
                outfile_name="best_MSA",
                combined_fastas=unaligned_seqs,
                prank_exe=prank_exe,
                add_args=args,
                logger=logger)
        else:
            raise ValueError("Construction of MSA skipped because " +
                             "%s is not a valid executable!", prank_exe)
    elif msa_tool == "mafft":
        if check_installed_tools(executable=mafft_exe,
                                 hard=False,
                                 logger=logger):
            msa_cmd, results_path = prepare_mafft_cmd(
                outdir=outdir,
                outfile_name="best_MSA.fasta",
                combined_fastas=unaligned_seqs,
                mafft_exe=mafft_exe,
                add_args=args,
                logger=logger)
        else:
            raise ValueError("Construction of MSA skipped because " +
                             "%s is not a valid executable!", mafft_exe)
    return(msa_cmd, results_path)


def plot_alignment_3d(consensus, tseq, output_prefix):
    """ under development.  no progress for a while
    """
    from mpl_toolkits.mplot3d import Axes3D
    import matplotlib
    import numpy as np
    from scipy.interpolate import interp1d
    from matplotlib import cm
    from matplotlib import pyplot as plt
    step = 0.04
    maxval = 1.0
    fig = plt.figure()
    ax = Axes3D(fig)
    ### Make differences

    # for index, value in enumerate(consensus):


    ###
    u=np.array([0,1,2,1,0,2,4,6,4,2,1])
    v=np.array([4,4,6,3,6,4,1,4,4,4,4])
    r=np.array([0,1,2,3,4,5,6,7,8,9,10])
    f=interp1d(r,u)

    # walk along the circle
    p = np.linspace(0,2*np.pi,20)
    R,P = np.meshgrid(r,p)
    # transform them to cartesian system
    X,Y = R*np.cos(P),R*np.sin(P)

    # Z=2
    # print(X)
    # print(Y)
    Z=f(R)

    # ax.plot_surface(X, Y, Z, rstride=1, cstride=1, cmap=cm.jet)
    ax.scatter(X, Y, Z)#, rstride=1, cstride=1, cmap=cm.jet)
    ax.set_xticks([])
    fig.savefig(str(output_prefix + '3d..png'), dpi=(200))
    fig.savefig(str(output_prefix + '3d..pdf'), dpi=(200))


def savitzky_golay(y, window_size, order, deriv=0, rate=1):
    r"""
    http://scipy.github.io/old-wiki/pages/Cookbook/SavitzkyGolay
    Smooth (and optionally differentiate) data with a Savitzky-Golay filter.
    The Savitzky-Golay filter removes high frequency noise from data.
    It has the advantage of preserving the original shape and
    features of the signal better than other types of filtering
    approaches, such as moving averages techniques.
    Parameters
    ----------
    y : array_like, shape (N,)
        the values of the time history of the signal.
    window_size : int
        the length of the window. Must be an odd integer number.
    order : int
        the order of the polynomial used in the filtering.
        Must be less then `window_size` - 1.
    deriv: int
        the order of the derivative to compute (default = 0 means only smoothing)
    Returns
    -------
    ys : ndarray, shape (N)
        the smoothed signal (or it's n-th derivative).
    Notes
    -----
    The Savitzky-Golay is a type of low-pass filter, particularly
    suited for smoothing noisy data. The main idea behind this
    approach is to make for each point a least-square fit with a
    polynomial of high order over a odd-sized window centered at
    the point.
    Examples
    --------
    t = np.linspace(-4, 4, 500)
    y = np.exp( -t**2 ) + np.random.normal(0, 0.05, t.shape)
    ysg = savitzky_golay(y, window_size=31, order=4)
    import matplotlib.pyplot as plt
    plt.plot(t, y, label='Noisy signal')
    plt.plot(t, np.exp(-t**2), 'k', lw=1.5, label='Original signal')
    plt.plot(t, ysg, 'r', label='Filtered signal')
    plt.legend()
    plt.show()
    References
    ----------
    .. [1] A. Savitzky, M. J. E. Golay, Smoothing and Differentiation of
       Data by Simplified Least Squares Procedures. Analytical
       Chemistry, 1964, 36 (8), pp 1627-1639.
    .. [2] Numerical Recipes 3rd Edition: The Art of Scientific Computing
       W.H. Press, S.A. Teukolsky, W.T. Vetterling, B.P. Flannery
       Cambridge University Press ISBN-13: 9780521880688
    """
    import numpy as np
    from math import factorial

    try:
        window_size = np.abs(np.int(window_size))
        order = np.abs(np.int(order))
    except ValueError:
        raise ValueError("window_size and order have to be of type int")
    if window_size % 2 != 1 or window_size < 1:
        raise TypeError("window_size size must be a positive odd number")
    if window_size < order + 2:
        raise TypeError("window_size is too small for the polynomials order")
    order_range = range(order + 1)
    half_window = (window_size - 1) // 2
    # precompute coefficients
    b = np.mat([[k ** i for i in order_range] for k in range(- half_window, half_window + 1)])
    m = np.linalg.pinv(b).A[deriv] * rate ** deriv * factorial(deriv)
    # pad the signal at the extremes with
    # values taken from the signal itself
    firstvals = y[0] - np.abs(y[1: half_window + 1][:: -1] - y[0] )
    lastvals = y[-1] + np.abs(y[-half_window - 1: -1][:: -1] - y[-1])
    y = np.concatenate((firstvals, y, lastvals))
    return np.convolve(m[::-1], y, mode='valid')


def main(clusters, genome_records, logger, verbose, no_revcomp,
         output, circular, flanking, prefix_name):
    get_rev_comp = no_revcomp is False  # kinda clunky
    extracted_regions = []
    logger.debug(clusters)
    for cluster in clusters:  # for each cluster of loci
        # get seq record that cluster is  from
        try:
            cluster.seq_record = \
                get_genbank_rec_from_multigb(recordID=cluster.sequence_id,
                                             genbank_records=genome_records)
        except Exception as e:
            logger.error(e)
            sys.exit(1)
        # make coord list
        try:
            extract_coords_from_locus(
                cluster=cluster, feature=cluster.feat_of_interest,
                logger=logger)
        except Exception as e:
            logger.error(e)
            sys.exit(1)
            logger.debug(
                "Here are the detected region,coords, strand, product, " +
                "locus tag, subfeatures and sequence id of the results:")
            logger.debug(str(cluster.__dict__))
        if circular:
            cluster_post_pad = pad_genbank_sequence(cluster=cluster,
                                                    logger=logger)
        else:
            cluster_post_pad = cluster

        #  given coords and a sequnce, extract the region as a SeqRecord
        try:
            cluster_post_stitch =\
                stitch_together_target_regions(cluster=cluster_post_pad,
                                               flanking=flanking,
                                               logger=logger,
                                               circular=circular,
                                               revcomp=get_rev_comp)
        except Exception as e:
            logger.error(e)
            sys.exit(1)
        extracted_regions.append(cluster_post_stitch.extractedSeqRecord)
    # after each cluster has been extracted, write out results
    logger.debug(extracted_regions)
    file_list = []
    for index, region in enumerate(extracted_regions):
        logger.debug(index)
        logger.debug(region)
        if prefix_name is None:
            prefix_name = date
        # make dir for blastable partial seeds
        graded_output = os.path.join(
            output,
            "{0}_region_{1}_{2}.fasta".format(
                prefix_name, index + 1, "graded_flanking_regions"))
        ####  make this a cmdline option, maybe
        do_blast = True
        if do_blast:
            os.makedirs(graded_output)
            for i in range(0, flanking + 100, 100):
                short_rec = SeqRecord(
                    #  needs the -1 to avoid the  index-by-negative-zero-error
                    Seq(str(region.seq[i: -i - 1]), IUPAC.IUPACAmbiguousDNA()),
                    id="{0}_{1}-bp_flanks".format(region.id, flanking - i))
                blast_fasta_filename = \
                    "{0}_region_{1}_{2}_{3}-bp_flanks.fasta".format(
                        prefix_name, index + 1, "riboSnag", flanking - i)
                with open(os.path.join(graded_output, blast_fasta_filename),
                          "w") as outfile_blast:
                    SeqIO.write(short_rec, outfile_blast, "fasta")
                file_list.append(
                    os.path.join(graded_output, blast_fasta_filename))
        ####
        filename = "{0}_region_{1}_{2}.fasta".format(
            prefix_name, index + 1, "riboSnag")
        # TODO make discription work when writing seqrecord
        # TODO move date tag to fasta description?
        # i.description = str("{0}_riboSnag_{1}_flanking_{2}_within".format(
        #                           output_index, args.flanking, args.within))
        with open(os.path.join(output, filename), "w") as outfile:
            SeqIO.write(region, outfile, "fasta")
            outfile.write('\n')
    # write out the whole file as a fasta as well...
    # append in case of multiple records
    ref_fasta = os.path.join(output, str(prefix_name + "_genome.fasta"))
    with open(ref_fasta, "a") as outfa:
        for record in genome_records:
            SeqIO.write(record, outfa, "fasta")
            outfa.write('\n')
    return extracted_regions, ref_fasta, file_list


def get_makeblastdb_cmd(input_file, input_type="fasta", dbtype="prot",
                        title="blastdb", out="blastdb",
                        makeblastdb_exe='', logger=None):
    """
    This runs make blast db with the given parameters
    requires logging, os, subprocess, shutil
    """
    if makeblastdb_exe == '':
        makeblastdb_exe = shutil.which("makeblastdb")
        if makeblastdb_exe is None:
            logger.error("error finding makeblastdb executable")
            raise ValueError("no executable for makeblastdb found!")
    makedbcmd = str("{0} -in {1} -input_type {2} -dbtype {3} " +
                    "-title {4} -out {5}").format(makeblastdb_exe,
                                                  input_file,
                                                  input_type,
                                                  dbtype, title, out)
    return makedbcmd


def make_blast_cmds(filename_list, blast_type, output, blastdb, date,
                    logger=None):
    """given a file, make a blast cmd, and return path to output csv
    """
    blast_cmds = []
    blast_outputs = []
    for f in filename_list:
        if logger:
            logger.debug("creating blast cmds for %s", f)
        output_path_tab = "{0}_{1}_results_{2}.tab".format(
            os.path.join(output, date),
            blast_type,
            os.path.basename(f))
        blast_cline = NcbiblastnCommandline(query=f,
                                            db=blastdb, evalue=10,
                                            outfmt=6, out=output_path_tab)
        if blast_type == 'blastn':
            add_params = str(
                " -num_threads 1 -max_target_seqs " +
                "2000 -task blastn -perc_identity 97")
        elif blast_type == 'dc_megablast':
            add_params = str(
                " -num_threads 1 -max_target_seqs " +
                "2000 -task dc_megablast -perc_identity 97")
        elif blast_type == 'tblastx':
            add_params = str(
                " -num_threads 1 -max_target_seqs 2000 " +
                "-query_gencode 11 -db_gencode 11 -perc_identity 97")
        else:
            raise ValueError("must use either blastn or tblastx")
        blast_command = str(str(blast_cline) + add_params)
        blast_cmds.append(blast_command)
        blast_outputs.append(output_path_tab)
    return(blast_cmds, blast_outputs)


def merge_outfiles(files, outfile_name, logger=None):
    """
    """
    # only grab .tab files, ie, the blast output
    logger.debug("files to merge:")
    logger.debug("\n".join(files))
    filelist = [i for i in files if "tab" in os.path.splitext(i)[-1]]
    if len(filelist) == 0:
        if logger:
            logger.error("No BLAST output files found!")
        raise FileNotFoundError

    elif len(filelist) == 1:
        if logger:
            logger.warning("only one file found! no merging needed")
        return(filelist)
    else:
        nfiles = len(filelist)
        if logger:
            logger.info("merging %i blast results to %s", nfiles, outfile_name)
        fout = open(outfile_name, "a")
        # first file:
        for line in open(filelist[0]):
            fout.write(line)
        #  now the rest, ignoring the header:
        for num in range(1, nfiles):
            f = open(filelist[num])
            for line in f:
                fout.write(line)
            f.close()  # not really needed
        fout.close()
    return(outfile_name)


def run_blast(query_list, ref, name, output, mbdb_exe='', logger=None):
    date = str(datetime.datetime.now().strftime('%Y%m%d'))
    logger.info("constructing makeblastdb command")
    db_name = os.path.join(output, str(name + "_db"), name)
    mbdb_cmd = get_makeblastdb_cmd(
        input_file=ref, input_type="fasta", dbtype="nucl",
        title=name, out=db_name,
        makeblastdb_exe=mbdb_exe, logger=logger)
    logger.debug(mbdb_cmd)
    logger.info("creating blast cmds")
    blast_outdir = os.path.join(output, "BLAST_results")
    os.makedirs(blast_outdir)
    blast_cmds, paths_to_outputs = make_blast_cmds(
        filename_list=query_list, blast_type="blastn",
        output=blast_outdir, blastdb=db_name, date=date, logger=logger)
    logger.info("Making BLAST Database")
    subprocess.run(mbdb_cmd,
                   shell=sys.platform != "win32",
                   stdout=subprocess.PIPE,
                   stderr=subprocess.PIPE,
                   check=True)
    # pool = multiprocessing.Pool(processes=args.cores)
    pool = multiprocessing.Pool()
    logger.info("Running BLAST commands")
    logger.debug("Running the following commands in parallel " +
                 "(this could take a while):")
    logger.debug("\n" + "\n".join([x for x in blast_cmds]))
    results = [
        pool.apply_async(subprocess.run,
                         (cmd,),
                         {"shell": sys.platform != "win32",
                          "stdout": subprocess.PIPE,
                          "stderr": subprocess.PIPE,
                          "check": True})
        for cmd in blast_cmds]
    pool.close()
    pool.join()
    reslist = []
    reslist.append([r.get() for r in results])
    logger.debug("merging resulting blast files")
    merged_tsv = merge_outfiles(
        paths_to_outputs,
        os.path.join(blast_outdir,
                     str(date + "_results_merged.csv")),
        logger=logger)
    logger.info("Merged BLAST results can be found here for perusal or plotting:")
    logger.info(merged_tsv)


if __name__ == "__main__":
    args = get_args()
    output_root = os.path.abspath(os.path.expanduser(args.output))
    # Create output directory only if it does not exist
    try:
        os.makedirs(args.output)
    except FileExistsError:
        # '#' is needed in case streaming output eventually
        print("#Selected output directory %s exists" %
              args.output)
        if not args.clobber:
            print("exiting")
            sys.exit(1)
        else:
            print("# continuing, and risking potential loss of data")
    log_path = os.path.join(output_root,
                            str("{0}_riboSnag_log.txt".format(
                                time.strftime("%Y%m%d%H%M"))))
    logger = set_up_logging(verbosity=args.verbosity,
                            outfile=log_path,
                            name=__name__)

    logger.debug("Usage:\n{0}\n".format(str(" ".join([x for x in sys.argv]))))
    logger.debug("All settings used:")
    for k, v in sorted(vars(args).items()):
        logger.debug("%s: %s", k, v)
    date = str(datetime.datetime.now().strftime('%Y%m%d'))
    # test whether executables are there
    executables = ['barrnap']
    test_ex = [check_installed_tools(x, logger=logger) for x in executables]
    if all(test_ex):
        logger.debug("All needed system executables found!")
        logger.debug(str([shutil.which(i) for i in executables]))
    check_version_from_cmd(
        exe='barrnap',
        cmd='', line=2,
        pattern=r"barrnap (?P<version>[^-]+)",
        where='stderr',
        logger=logger,
        coerce_two_digit=True,
        min_version="0.0.7")
    # parse cluster file
    try:
        clusters = parse_clustered_loci_file(args.clustered_loci,
                                             gb_filepath=args.genbank_genome,
                                             output_root='',
                                             padding=args.padding,
                                             circular=args.circular,
                                             logger=logger)
    except Exception as e:
        logger.error(e)
        sys.exit(1)
    # parse genbank records
    with open(args.genbank_genome) as fh:
        genome_records = list(SeqIO.parse(fh, 'genbank'))
    regions = []
    logger.info("clusters:")
    logger.info(clusters)
    if args.name is None:  # if none given, use date for name (for out files)
        args.name = date
    regions, ref_fasta, region_files = main(
        clusters=clusters,
        genome_records=genome_records,
        logger=logger,
        verbose=False,
        flanking=args.flanking,
        output=output_root,
        circular=args.circular,
        prefix_name=args.name,
        no_revcomp=args.no_revcomp,
    )

    # make MSA and calculate entropy
    if not args.skip_check:
        if args.clobber:
            logger.error("Cannot safely check SMA when --clobber is used!")
            sys.exit(1)

        unaligned_seqs = combine_contigs(contigs_dir=output_root,
                                         pattern="*riboSnag*",
                                         contigs_name="riboSnag_unaligned",
                                         ext=".fasta", verbose=False,
                                         logger=logger)
        msa_cmd, results_path = make_msa(msa_tool=args.msa_tool,
                                         unaligned_seqs=unaligned_seqs,
                                         prank_exe=args.prank_exe,
                                         args='',
                                         mafft_exe=args.mafft_exe,
                                         outdir=output_root,
                                         logger=logger)

        logger.info("Running %s for MSA", args.msa_tool)
        subprocess.run(msa_cmd,
                       shell=sys.platform != "win32",
                       stdout=subprocess.PIPE,
                       stderr=subprocess.PIPE,
                       check=True)

        seq_entropy, names, tseq = calc_entropy_msa(results_path)
        if args.msa_kmers:
            with open(results_path, 'r') as resfile:
                kmer_seqs = list(SeqIO.parse(resfile, "fasta"))
        else:
            kmer_seqs = regions
        counts, names = profile_kmer_occurances(kmer_seqs,
                                                # alph='atcg-',
                                                k=5,
                                                logger=logger)

        mca_df = plot_pairwise_least_squares(
            counts=counts, names_list=names,
            output_prefix=os.path.join(
                output_root,
                "sum_least_squares"))
        gff, consensus_cov, annos = annotate_msa_conensus(
            tseq_array=tseq,
            pattern='product=(.+?)$',
            seq_file=os.path.join(
                output_root,
                "test_consensus.fasta"),
            barrnap_exe=args.barrnap_exe,
            countdashcov=False,   # include -'s in coverage
            collapseNs=False,  # include -'s in consensus
            excludedash=False,
            kingdom=args.kingdom,
            logger=logger)
        title = str("Shannon Entropy by Position\n" +
                    os.path.basename(
                        os.path.splitext(
                            args.genbank_genome)[0]))

        return_code = plot_scatter_with_anno(
            data=seq_entropy,
            consensus_cov=consensus_cov,
            names=["Position", "Entropy"],
            title=title,
            anno_list=annos,
            output_prefix=os.path.join(
                output_root,
                "entropy_plot"))
        # plot_alignment_3d(
        #     output_prefix=os.path.join(output_root, "entropy_plot"),
        #     consensus=consensus_cov,
        #     tseq=tseq)
    run_blast(query_list=region_files, ref=ref_fasta,
              mbdb_exe=args.makeblastdb_exe, name=args.name,
              output=args.output, logger=logger)



# -*- coding: utf-8 -*-
"""
Created on Tue Aug 30 08:57:31 2016
@author: nicholas

"""
import sys
import logging
import os
import unittest

from riboSeed.riboSelect import get_filtered_locus_tag_dict, \
    pure_python_kmeans
from pyutilsnrw.utils3_5 import get_genbank_record, check_installed_tools

logger = logging
sys.dont_write_bytecode = True


@unittest.skipIf((sys.version_info[0] != 3) or (sys.version_info[1] < 5),
                 "Subprocess.call among otherthings wont run if you try this" +
                 " with less than python 3.5")
class riboSelect_TestCase(unittest.TestCase):
    """ still need to write tests for  count_feature_hits
    """
    def setUp(self):
        self.curdir = os.getcwd()
        self.testdirname = os.path.join(os.path.dirname(__file__),
                                        "output_utils3_5_tests")
        self.test_loci_file = os.path.join(os.path.dirname(__file__),
                                           str("references" + os.path.sep +
                                               'grouped_loci_reference.txt'))
        self.test_gb_file = os.path.join(os.path.dirname(__file__),
                                         str("references" + os.path.sep +
                                             'NC_011751.1.gb'))
        self.test_loci_file = os.path.join(os.path.dirname(__file__),
                                           str("references" + os.path.sep +
                                               'grouped_loci_reference.txt'))
        if not os.path.exists(self.testdirname):
            os.makedirs(self.testdirname, exist_ok=True)

    def test_locus_tag_dict(self):
        """test different scenarios of trying to get features
        from a gb file with get_filtered_locus_tag_dict
        """
        # test single genbank record
        record = get_genbank_record(self.test_gb_file)[0]
        # with self.assertRaises(SystemExit):
        # test that when given a bad feature empty results are returned
        bum_coords, bum_nfeat, bum_nfeat_simple = \
            get_filtered_locus_tag_dict([record], feature="RRNA",
                                        specific_features="16S",
                                        verbose=False,
                                        logger=logger)
        self.assertTrue([len(x) == 0 for x in
                         [bum_nfeat, bum_nfeat_simple, bum_coords]])
        bum_coords2, bum_nfeat2, bum_nfeat_simple2 = \
            get_filtered_locus_tag_dict([record], feature="rRNA",
                                        specific_features="18S",
                                        verbose=False,
                                        logger=logger)
        self.assertTrue([len(x) == 0 for x in
                         [bum_nfeat2, bum_nfeat_simple2, bum_coords2]])

        filtered, nfeat, nfeat_simple = \
            get_filtered_locus_tag_dict([record], feature="rRNA",
                                        specific_features="16S",
                                        verbose=False,
                                        logger=logger)
        # for k, v in filtered.items():
        #     print(k, v)
        self.assertEqual(len(filtered), 7)
        # check length of 1st item (ie, occurances) is same
        self.assertEqual(len(nfeat['NC_011751.1']),
                         len(nfeat_simple['NC_011751.1']))
        self.assertEqual([8614, 'NC_011751.1', 'ECUMN_16S_4', 'rRNA',
                          ['ribosomal', 'RNA', '16S']],
                         filtered[('NC_011751.1', 4428675)])
        self.assertEqual(nfeat_simple, {'NC_011751.1': [7]})

    def test_dopey_kmeans_function(self):
        """As they say, don't look a gift kmeans in the mouth
        """
        r_is_installed = check_installed_tools("R",
                                               hard=False, logger=logger)
        if r_is_installed:
            test_for_clustering = [4, 5, 5, 6, 10,
                                   3, 18, 34, 44, 38]
            # debug is true in case we need to assess created files
            clusters = pure_python_kmeans(test_for_clustering,
                                          centers=3, DEBUG=True,
                                          kind=int)
            ref_dict = {'2': [3, 4, 5, 5, 6, 10], '1': [34, 38, 44], '3': [18]}
            self.assertEqual(ref_dict, clusters)
            # delete list.csv and Kmeans.R if we get this far
            os.remove(os.path.join(os.getcwd(),
                                   "pure_python_kmeans_list.csv"))
            os.remove(os.path.join(os.getcwd(), "km_script.R"))

    def tearDown(self):
        pass


if __name__ == '__main__':
    unittest.main()




import cProfile
import unittest
import pstats

if __name__ == '__main__':
   suite = unittest.TestLoader().discover('.')
   def runtests():
      unittest.TextTestRunner().run(suite)
   s = cProfile.run('runtests()',sort='cumtime')



# -*- coding: utf-8 -*-
"""
Created on Tue Aug 30 08:57:31 2016
@author: nicholas

"""
import sys
import logging
import shutil
# import subprocess
import os
import unittest
# import multiprocessing

# from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from argparse import Namespace

# I hate this line but it works :(
sys.path.append(os.path.join(
    os.path.dirname(os.path.dirname(__file__)), "riboSeed"))


from pyutilsnrw.utils3_5 import md5, file_len, copy_file, get_number_mapped

from riboSeed.riboSeed import SeedGenome, NgsLib, LociMapping, Exes, \
    map_to_genome_ref_smalt, map_to_genome_ref_bwa, \
    add_coords_to_clusters, partition_mapping, \
    convert_bam_to_fastqs_cmd, get_smalt_full_install_cmds,\
    generate_spades_cmd, estimate_distances_smalt, get_final_assemblies_cmds,\
    nonify_empty_lib_files, make_faux_genome, \
    evaluate_spades_success, prepare_next_mapping, make_mapped_partition_cmds,\
    make_unmapped_partition_cmds, make_quick_quast_table, \
    check_fastqs_len_equal, parse_subassembly_return_code, \
    make_spades_empty_check, get_samtools_depths

from riboSeed.riboSnag import parse_clustered_loci_file, \
    extract_coords_from_locus, stitch_together_target_regions, \
    pad_genbank_sequence, prepare_prank_cmd, prepare_mafft_cmd, \
    calc_Shannon_entropy, plot_scatter_with_anno, \
    profile_kmer_occurances, plot_pairwise_least_squares, make_msa


sys.dont_write_bytecode = True

logger = logging


@unittest.skipIf((sys.version_info[0] != 3) or (sys.version_info[1] < 5),
                 "Subprocess.call among other things wont run if tried " +
                 " with less than python 3.5")
class riboSeedTestCase(unittest.TestCase):
    """ tests for riboSeed.py
    """
    def setUp(self):
        self.test_dir = os.path.join(os.path.dirname(__file__),
                                     "output_riboseed2_tests")
        self.spades_dir = os.path.join(os.path.dirname(__file__),
                                       "output_riboseed_tests",
                                       "SPAdes_results")
        self.ref_dir = os.path.join(os.path.dirname(__file__), "references")
        self.ref_gb = os.path.join(self.ref_dir,
                                   'NC_011751.1.gb')
        self.ref_fasta = os.path.join(self.test_dir,
                                      'cluster1.fasta')
        self.good_contig = os.path.join(self.ref_dir,
                                        'contigs.fasta')
        self.short_contig = os.path.join(self.ref_dir,
                                         'contigs.fasta')
        self.ref_Ffastq = os.path.join(self.ref_dir,
                                       'toy_reads1.fq')
        self.ref_Rfastq = os.path.join(self.ref_dir,
                                       'toy_reads2.fq')
        self.ref_bam_prefix = os.path.join(self.ref_dir,
                                           'test_bam_to_fastq')
        self.smalt_exe = "smalt"
        self.bwa_exe = "bwa"
        self.samtools_exe = "samtools"
        self.spades_exe = "spades.py"
        self.quast_exe = "quast.py"
        self.python2_7_exe = "python2"
        self.test_estimation_file = os.path.join(self.test_dir,
                                                 "est_distance.sam")
        self.map_results_prefix = os.path.join(self.test_dir,
                                               "test_mapping")
        self.fastq_results_prefix = os.path.join(self.test_dir,
                                                 "test_bam_to_fastq")
        self.test_loci_file = os.path.join(os.path.dirname(__file__),
                                           str("references" + os.path.sep +
                                               'grouped_loci_reference.txt'))
        self.args = Namespace(skip_contol=False, kmers="21,33,55,77,99",
                              spades_exe="spades.py",
                              quast_exe="python2.7 quast.py",
                              cores=2)
        self.cores = 2
        self.maxDiff = 2000
        self.to_be_removed = []
        if not os.path.exists(self.test_dir):
            os.makedirs(self.test_dir, exist_ok=True)
        self.copy_fasta()

    def copy_fasta(self):
        """ make a disposable copy
        """
        copy_file(os.path.join(self.ref_dir, 'cluster1.fasta'), os.path.dirname(self.ref_fasta),
                  name="cluster1.fasta")
        self.to_be_removed.append(self.ref_fasta)

    def test_print_headsup(self):
        print("""
        Thanks for running these tests!! You will see several logger warning
        messages that you should IGNORE. **sigh of relief** They verify error
        reporting in the actual function, and cannot be removed. You should
        see the following errors if all went properly:
        \t WARNING:root:The first iteration's assembly's best contig is not greater than length set by --min_assembly_len. Assembly will likely fail if the contig does not meet the length of the seed
        \t WARNING:root:Continuing, but if this occurs for more than one seed, we reccommend  you abort and retry with longer seeds, a different ref, or re-examine the riboSnag clustering
        \t WARNING:root:The first iteration's assembly's best contig is not greater than length set by --min_assembly_len. Assembly will likely fail if the contig does not meet the length of the seed
        \t WARNING:root:SEED_cluster-7-iter1: No output from SPAdes this time around...
        \t WARNING:root:read file readR is empty and will not be used for mapping!
        """)

    def test_make_spades_empty_check(self):
        lib1 = "/path/to/lib1.fastq"
        lib2 = "/path/to/lib2.fastq"
        lib3 = "/path/to/lib3.fastq"
        cmd = "spades command"
        fullcmd = make_spades_empty_check(
            liblist=[lib1, lib2, lib3],
            cmd=cmd,
            logger=logger)
        self.assertEqual(
            str("if [ -s {0} ] && [ -s {1} ] && [ -s {2} ] ; " +
                "then {3} ; else echo 'input lib not found, " +
                "skipping this SPAdes call' ; fi").format(
                    lib1, lib2, lib3, cmd),
            fullcmd)

    def test_Exes_bad_method(self):
        """check badd method arg"""
        with self.assertRaises(ValueError):
            Exes(samtools=self.samtools_exe,
                 quast=self.quast_exe,
                 smalt=self.smalt_exe,
                 python2_7=self.python2_7_exe,
                 spades=self.spades_exe,
                 bwa=self.bwa_exe,
                 method="bowtie")

    def test_Exes_bad_attribute(self):
        """ check bad instatiation """
        with self.assertRaises(AssertionError):
            Exes(samtools=None,
                 quast=self.quast_exe,
                 spades=self.spades_exe,
                 python2_7=self.python2_7_exe,
                 smalt=self.smalt_exe,
                 bwa=self.bwa_exe,
                 method="bwa")

    def test_Exes_bad_exe(self):
        """check with nonexistant executable"""
        with self.assertRaises(ValueError):
            Exes(samtools="nottheactualsamtools_exe",
                 quast=self.quast_exe,
                 python2_7=self.python2_7_exe,
                 smalt=self.smalt_exe,
                 spades=self.spades_exe,
                 bwa=self.bwa_exe,
                 method="bwa")

    @unittest.skipIf(shutil.which("bwa") is None or
                     shutil.which("quast.py") is None or
                     shutil.which("smalt") is None or
                     shutil.which("python2.7") is None or
                     shutil.which("spades.py") is None,
                     "bwa executable not found, skipping.If this isnt an " +
                     "error from travis deployment, you probably "+
                     "should install it")
    def test_Exes_(self):
        """check with  executable"""
        test_exes = Exes(samtools=self.samtools_exe,
                         quast=self.quast_exe,
                         python2_7=self.python2_7_exe,
                         smalt=self.smalt_exe,
                         spades=self.spades_exe,
                         bwa=self.bwa_exe,
                         method="bwa")
        self.assertEqual(test_exes.mapper, shutil.which(test_exes.bwa))

    def test_NgsLib(self):
        # make a non-master object
        testlib_pe_s = NgsLib(
            name="test",
            master=False,
            readF=self.ref_Ffastq,
            readR=self.ref_Rfastq,
            readS0="dummy",
            ref_fasta=self.ref_fasta,
            mapper_exe=self.smalt_exe)
        #  this test should go away when we allow running with single lib
        with self.assertRaises(ValueError):
            testlib_s = NgsLib(
                name="test",
                master=False,
                readF=None,
                readR=None,
                readS0=self.ref_Ffastq,
                ref_fasta=self.ref_fasta,
                mapper_exe=self.smalt_exe)
            self.assertEqual(testlib_s.libtype, "s_1")
            self.assertEqual(testlib_s.readlen, 145.0)
        # test unnamed fails
        with self.assertRaises(ValueError):
            NgsLib(
                name=None,
                master=False,
                readF=self.ref_Ffastq,
                readR=self.ref_Rfastq,
                readS0="dummy",
                ref_fasta=self.ref_fasta,
                mapper_exe=self.smalt_exe)
        # self.assertFalse(os.path.exists(os.path.join(
        #     self.test_dir, "smalt_distance_est.sam")))
        self.assertEqual(testlib_pe_s.libtype, "pe_s")
        self.assertEqual(testlib_pe_s.readlen, None)
        # test fails with singe PE file
        with self.assertRaises(ValueError):
            NgsLib(
                name=None,
                master=False,
                readF=self.ref_Ffastq,
                readR=None,
                readS0="dummy",
                ref_fasta=self.ref_fasta,
                mapper_exe=self.smalt_exe)

        # # check master (ie, generate a distance file with smalt
        # testlib_pe = NgsLib(
        #     name="test",
        #     master=True,
        #     make_dist=True,
        #     readF=self.ref_Ffastq,
        #     readR=self.ref_Rfastq,
        #     ref_fasta=self.ref_fasta,
        #     mapper_exe=self.smalt_exe,
        #     logger=logger)
        # self.assertTrue(os.path.exists(os.path.join(
        #     self.test_dir, "smalt_distance_est.sam")))
        # self.assertEqual(testlib_pe.libtype, "pe")
        # self.assertEqual(testlib_pe.readlen, 145.0)
        # self.to_be_removed.append(testlib_pe.smalt_dist_path)

    def test_SeedGenome(self):
        gen = SeedGenome(
            max_iterations=2,
            clustered_loci_txt=self.test_loci_file,
            genbank_path=self.ref_gb,
            loci_clusters=None,
            output_root=self.test_dir)
        self.assertTrue(os.path.exists(os.path.join(self.test_dir,
                                                    "NC_011751.1.fasta")))
        self.assertTrue(os.path.exists(self.test_dir))
        self.assertTrue(isinstance(gen.seq_records[0], SeqRecord))

    def test_SeedGenome_bad_instatiation(self):
        with self.assertRaises(ValueError):
            SeedGenome(
                max_iterations=2,
                genbank_path=self.ref_gb,
                clustered_loci_txt=self.test_loci_file,
                loci_clusters=None,
                output_root=None)

    def test_LociMapping(self):
        with self.assertRaises(ValueError):
            LociMapping(
                name=None,
                iteration=1,
                mapping_subdir=os.path.join(self.test_dir, "LociMapping"))
        testmapping = LociMapping(
            name="test",
            iteration=1,
            mapping_subdir=os.path.join(self.test_dir, "LociMapping"))
        self.assertTrue(os.path.isdir(testmapping.mapping_subdir))

    def test_get_smalt_full_install_cmds(self):
        """
        """
        smalttestdir = os.path.join(os.path.dirname(os.path.dirname(__file__)),
                                    "sample_data",
                                    "smalt_test", "")
        ref = os.path.join(smalttestdir, "ref_to_test_bambamc.fasta")
        index = os.path.join(smalttestdir, "test_index")
        test_bam = os.path.join(smalttestdir, "test_mapping.bam")
        test_reads = os.path.join(smalttestdir, "reads_to_test_bambamc.fastq")

        check_cmds = get_smalt_full_install_cmds(self.smalt_exe, logger=logger)
        check_cmds_ref = [
            "{0} index {1} {2}".format(
                self.smalt_exe, index, ref),
            "{0} map -f bam -o {1} {2} {3}".format(
                self.smalt_exe, test_bam, index, test_reads)]
        for index, cmd in enumerate(check_cmds):
            self.assertEqual(cmd, check_cmds_ref[index])

    @unittest.skipIf(shutil.which("smalt") is None, "smat executable not found, skipping."+
                     "If this isnt an error from travis deployment, you probably "+
                     "should install it")
    def test_estimate_distances_smalt(self):
        """ test estimate insert disances
        """
        if os.path.exists(self.test_estimation_file):
            print("warning! existing distance estimation file!")
        est_file = estimate_distances_smalt(outfile=self.test_estimation_file,
                                            smalt_exe=self.smalt_exe,
                                            ref_genome=self.ref_fasta,
                                            fastq1=self.ref_Ffastq,
                                            fastq2=self.ref_Rfastq,
                                            cores=1,
                                            logger=logger)
        ref_smi_md5 = "a444ccbcb486a8af29736028640e87cf"  # determined manually
        ref_sma_md5 = "4ce0c8b453f2bdabd73eaf8b5ee4f376"  # determined manually
        ref_mapping_len = 9271  # mapping doesnt have exact order, so cant md5
        self.assertEqual(ref_smi_md5, md5(str(est_file + ".smi")))
        self.assertEqual(ref_sma_md5, md5(str(est_file + ".sma")))
        self.assertEqual(ref_mapping_len, file_len(est_file))

    def test_add_coords_to_SeedGenome(self):
        gen = SeedGenome(
            max_iterations=1,
            genbank_path=self.ref_gb,
            clustered_loci_txt=self.test_loci_file,
            output_root=self.test_dir,
            logger=logger)
        gen.loci_clusters = parse_clustered_loci_file(
            filepath=gen.clustered_loci_txt,
            gb_filepath=gen.genbank_path,
            output_root=self.test_dir,
            padding=100,
            circular=False,
            logger=logger)
        add_coords_to_clusters(seedGenome=gen, logger=logger)
        self.assertEqual(
            gen.loci_clusters[0].loci_list[0].start_coord, 4656045)
        self.assertEqual(
            gen.loci_clusters[0].loci_list[0].end_coord, 4657586)

    def test_convert_bam_to_fastqs_cmd(self):
        testmapping = LociMapping(
            name="test",
            iteration=1,
            mapping_subdir=os.path.join(self.test_dir, "LociMapping"))
        cmd, ngs = convert_bam_to_fastqs_cmd(mapping_ob=testmapping,
                                             ref_fasta="test_reference.fasta",
                                             samtools_exe=self.samtools_exe,
                                             which='mapped',
                                             source_ext="_bam",
                                             logger=logger)
        cmd_ref = "{0} fastq {1} -1 {2} -2 {3} -s {4}".format(
            self.samtools_exe, testmapping.mapped_bam,
            os.path.join(testmapping.mapping_subdir,
                         "test_iteration_1_mappedreadF.fastq"),
            os.path.join(testmapping.mapping_subdir,
                         "test_iteration_1_mappedreadR.fastq"),
            os.path.join(testmapping.mapping_subdir,
                         "test_iteration_1_mappedreadS.fastq"), )
        self.assertEqual(cmd, cmd_ref)

    def test_generate_spades_cmds(self):
        """Question: why the heck is there a space before the -o?
        I troubleshot this for two hours to no avail.  Next!
        """
        testmapping = LociMapping(
            name="test",
            iteration=1,
            assembly_subdir=self.test_dir,
            ref_fasta=self.ref_fasta,
            mapping_subdir=os.path.join(self.test_dir, "LociMapping"))
        testngs1 = NgsLib(
            name="test",
            master=False,
            readF=self.ref_Ffastq,
            readR=self.ref_Rfastq,
            ref_fasta=self.ref_fasta,
            mapper_exe=self.smalt_exe)
        testngs2 = NgsLib(
            name="test",
            master=False,
            readF=self.ref_Ffastq,
            readR=self.ref_Rfastq,
            readS0=self.ref_Rfastq,
            ref_fasta=self.ref_fasta,
            mapper_exe=self.smalt_exe)
        # PE reads, prelim
        cmd1 = generate_spades_cmd(mapping_ob=testmapping, ngs_ob=testngs1,
                                   ref_as_contig='trusted',
                                   as_paired=True, addLibs="",
                                   prelim=True,
                                   k="21,33,55,77,99",
                                   spades_exe="spades.py",
                                   logger=logger)
        # PE with singletons
        cmd2 = generate_spades_cmd(mapping_ob=testmapping, ngs_ob=testngs2,
                                   ref_as_contig='trusted',
                                   as_paired=True, addLibs="",
                                   prelim=False,
                                   k="21,33,55,77,99",
                                   spades_exe="spades.py",
                                   logger=logger)
        # PE as sngle libraries
        cmd3 = generate_spades_cmd(mapping_ob=testmapping, ngs_ob=testngs1,
                                   ref_as_contig='trusted',
                                   as_paired=False, addLibs="",
                                   prelim=False,
                                   k="21,33,55,77,99",
                                   spades_exe="spades.py",
                                   logger=logger)
        # PE and singletons all as single libraries
        cmd4 = generate_spades_cmd(mapping_ob=testmapping, ngs_ob=testngs2,
                                   ref_as_contig='trusted',
                                   as_paired=False, addLibs="",
                                   prelim=False,
                                   k="21,33,55,77,99",
                                   spades_exe="spades.py",
                                   logger=logger)
        cmd1_ref = str(
            "spades.py --only-assembler --cov-cutoff off --sc --careful " +
            "-k 21,33,55,77,99 --pe1-1 {0} " +
            "--pe1-2 {1} --trusted-contigs {2}  -o {3}"
        ).format(
            self.ref_Ffastq, self.ref_Rfastq, self.ref_fasta, self.test_dir)
        self.assertEqual(cmd1, cmd1_ref)

        cmd2_ref = str(
            "spades.py --careful -k 21,33,55,77,99 --pe1-1 {0} " +
            "--pe1-2 {1} --pe1-s {2} --trusted-contigs {3}  -o {4}"
        ).format(
            self.ref_Ffastq, self.ref_Rfastq, self.ref_Rfastq,
            self.ref_fasta, self.test_dir)
        self.assertEqual(cmd2, cmd2_ref)

        cmd3_ref = str(
            "spades.py --careful -k 21,33,55,77,99 --pe1-s {0} " +
            "--pe2-s {1} --trusted-contigs {2}  -o {3}"
        ).format(
            self.ref_Ffastq, self.ref_Rfastq, self.ref_fasta, self.test_dir)
        self.assertEqual(cmd3, cmd3_ref)

        cmd4_ref = str(
            "spades.py --careful -k 21,33,55,77,99 --pe1-s {0} " +
            "--pe2-s {1} --pe3-s {1}  --trusted-contigs {2}  -o {3}"
        ).format(
            self.ref_Ffastq, self.ref_Rfastq, self.ref_fasta, self.test_dir)
        self.assertEqual(cmd4, cmd4_ref)

    def test_lib_check(self):
        empty_file = os.path.join(self.test_dir, "test_not_real_file")
        # make an empty file
        with open(empty_file, 'w') as ef:
            pass
        ngs_ob = NgsLib(
            name="test",
            master=False,
            readF=self.ref_Ffastq,
            readR=empty_file,
            ref_fasta=self.ref_fasta,
            mapper_exe=self.smalt_exe)
        nonify_empty_lib_files(ngsLib=ngs_ob, logger=logger)
        self.assertTrue(ngs_ob.readR is None)
        self.to_be_removed.append(empty_file)

    @unittest.skipIf(shutil.which("smalt") is None, "smat executable not found, skipping."+
                     "If this isnt an error from travis deployment, you probably "+
                     "should install it")
    def test_map_with_smalt(self):
        # becasue multiple mapping are assingmed randomly (pseudorandomly?),
        # this accepts a range of expected results
        testmapping = LociMapping(
            name="test",
            iteration=1,
            assembly_subdir=self.test_dir,
            ref_fasta=self.ref_fasta,
            mapping_subdir=os.path.join(self.test_dir, "LociMapping"))
        testngs = NgsLib(
            name="test",
            master=True,
            make_dist=True,
            readF=self.ref_Ffastq,
            readR=self.ref_Rfastq,
            ref_fasta=self.ref_fasta,
            mapper_exe=self.smalt_exe,
            logger=logger)

        map_to_genome_ref_smalt(mapping_ob=testmapping, ngsLib=testngs,
                                cores=4, samtools_exe=self.samtools_exe,
                                smalt_exe=self.smalt_exe, score_minimum=48,
                                scoring="match=1,subst=-4,gapopen=-4,gapext=-3",
                                step=3, k=5, logger=logger)
        mapped_str = get_number_mapped(testmapping.pe_map_bam,
                                       samtools_exe=self.samtools_exe)
        nmapped = int(mapped_str[0:5])
        nperc = float(mapped_str[-13:-8])
        print("\nSMALT: number of PE reads mapped: %f2" % nmapped)
        print("SMALT: percentage of PE reads mapped: %2f" % nperc)
        ###
        testngs2 = NgsLib(
            name="test",
            master=True,
            make_dist=True,
            readF=self.ref_Ffastq,
            readR=self.ref_Rfastq,
            readS0=self.ref_Rfastq,
            ref_fasta=self.ref_fasta,
            mapper_exe=self.smalt_exe,
            logger=logger)

        map_to_genome_ref_smalt(
            mapping_ob=testmapping, ngsLib=testngs2,
            cores=4, samtools_exe=self.samtools_exe,
            smalt_exe=self.smalt_exe, score_minimum=48,
            scoring="match=1,subst=-4,gapopen=-4,gapext=-3",
            step=3, k=5, logger=logger)
        mapped_str2 = get_number_mapped(testmapping.pe_map_bam,
                                        samtools_exe=self.samtools_exe)
        mapped_str2 = get_number_mapped(testmapping.s_map_bam,
                                        samtools_exe=self.samtools_exe)
        nmapped2 = int(mapped_str2[0:5])
        nperc2 = float(mapped_str2[-13:-8])
        print("SMALT: number of s reads mapped: %f2" % nmapped2)
        print("SMALT: percentage of s reads mapped: %f2" % nperc2)

        ###
        self.assertTrue(12500 < nmapped < 12680)
        self.assertTrue(34.3 < nperc < 34.81)
        ###
        self.assertTrue(6400 < nmapped2 < 6500)
        self.assertTrue(34.3 < nperc2 < 35.8)

    def test_parse_subassembly_return_code_0(self):
        gen = SeedGenome(
            max_iterations=1,
            genbank_path=self.ref_gb,
            clustered_loci_txt=self.test_loci_file,
            output_root=self.test_dir,
            logger=logger)
        gen.loci_clusters = parse_clustered_loci_file(
            filepath=gen.clustered_loci_txt,
            gb_filepath=gen.genbank_path,
            output_root=self.test_dir,
            padding=100,
            circular=False,
            logger=logger)
        gen.loci_clusters[0].assembly_success = 0
        parse_subassembly_return_code(cluster=gen.loci_clusters[0],
                                      logger=logger)
        self.assertTrue(gen.loci_clusters[0].continue_iterating)
        self.assertTrue(gen.loci_clusters[0].keep_contigs)

    def test_parse_subassembly_return_code_1(self):
        gen = SeedGenome(
            max_iterations=1,
            genbank_path=self.ref_gb,
            clustered_loci_txt=self.test_loci_file,
            output_root=self.test_dir,
            logger=logger)
        gen.loci_clusters = parse_clustered_loci_file(
            filepath=gen.clustered_loci_txt,
            gb_filepath=gen.genbank_path,
            output_root=self.test_dir,
            padding=100,
            circular=False,
            logger=logger)
        gen.loci_clusters[0].assembly_success = 1
        parse_subassembly_return_code(cluster=gen.loci_clusters[0],
                                      logger=logger)
        self.assertFalse(gen.loci_clusters[0].continue_iterating)
        self.assertFalse(gen.loci_clusters[0].keep_contigs)

    def test_parse_subassembly_return_code_2(self):
        gen = SeedGenome(
            max_iterations=1,
            genbank_path=self.ref_gb,
            clustered_loci_txt=self.test_loci_file,
            output_root=self.test_dir,
            logger=logger)
        gen.loci_clusters = parse_clustered_loci_file(
            filepath=gen.clustered_loci_txt,
            gb_filepath=gen.genbank_path,
            output_root=self.test_dir,
            padding=100,
            circular=False,
            logger=logger)
        gen.loci_clusters[0].assembly_success = 2
        parse_subassembly_return_code(cluster=gen.loci_clusters[0],
                                      logger=logger)
        self.assertFalse(gen.loci_clusters[0].continue_iterating)
        self.assertFalse(gen.loci_clusters[0].keep_contigs)

    def test_parse_subassembly_return_code_3(self):
        gen = SeedGenome(
            max_iterations=1,
            genbank_path=self.ref_gb,
            clustered_loci_txt=self.test_loci_file,
            output_root=self.test_dir,
            logger=logger)
        gen.loci_clusters = parse_clustered_loci_file(
            filepath=gen.clustered_loci_txt,
            gb_filepath=gen.genbank_path,
            output_root=self.test_dir,
            padding=100,
            circular=False,
            logger=logger)
        gen.loci_clusters[0].assembly_success = 3
        parse_subassembly_return_code(cluster=gen.loci_clusters[0],
                                      logger=logger)
        self.assertFalse(gen.loci_clusters[0].continue_iterating)
        self.assertFalse(gen.loci_clusters[0].keep_contigs)

    @unittest.skipIf(shutil.which("bwa") is None, "bwa executable not found, skipping."+
                     "If this isnt an error from travis deployment, you probably "+
                     "should install it")
    def test_map_with_bwa(self):
        # becasue multiple mapping are assingmed randomly (pseudorandomly?),
        # this accepts a range of expected results
        testmapping = LociMapping(
            name="test",
            iteration=1,
            assembly_subdir=self.test_dir,
            ref_fasta=self.ref_fasta,
            mapping_subdir=os.path.join(self.test_dir, "LociMapping"))
        testngs = NgsLib(
            name="test",
            master=True,
            readF=self.ref_Ffastq,
            readR=self.ref_Rfastq,
            ref_fasta=self.ref_fasta,
            mapper_exe=self.bwa_exe,
            logger=logger)

        map_to_genome_ref_bwa(mapping_ob=testmapping, ngsLib=testngs,
                              cores=4, samtools_exe=self.samtools_exe,
                              bwa_exe=self.bwa_exe, score_minimum=20,
                              logger=logger)
        mapped_str = get_number_mapped(testmapping.pe_map_bam,
                                       samtools_exe=self.samtools_exe)
        nmapped = int(mapped_str[0:5])
        nperc = float(mapped_str[-13:-8])
        print("\nBWA: number of PE reads mapped: %f2" % nmapped)
        print("BWA: percentage of PE reads mapped: %2f" % nperc)
        ###
        testngs2 = NgsLib(
            name="test",
            master=True,
            readF=self.ref_Ffastq,
            readR=self.ref_Rfastq,
            readS0=self.ref_Rfastq,
            ref_fasta=self.ref_fasta,
            mapper_exe=self.bwa_exe,
            logger=logger)

        map_to_genome_ref_bwa(mapping_ob=testmapping, ngsLib=testngs2,
                              cores=4, samtools_exe=self.samtools_exe,
                              bwa_exe=self.bwa_exe, score_minimum=20,
                              logger=logger)
        mapped_str2 = get_number_mapped(testmapping.pe_map_bam,
                                        samtools_exe=self.samtools_exe)
        mapped_str2 = get_number_mapped(testmapping.s_map_bam,
                                        samtools_exe=self.samtools_exe)
        nmapped2 = int(mapped_str2[0:5])
        nperc2 = float(mapped_str2[-13:-8])
        print("BWA: number of s reads mapped: %f2" % nmapped2)
        print("BWA: percentage of s reads mapped: %f2" % nperc2)

        ###
        self.assertTrue(13000 < nmapped < 14000)
        self.assertTrue(35.8 < nperc < 37.4)
        ###
        self.assertTrue(6300 < nmapped2 < 7000)
        self.assertTrue(35.0 < nperc2 < 36.8)

    def test_get_samtools_depths(self):
        """test bam generated by using SMALT to map the
        sample data reads (from the bam install check) to this reference
        (e coli genome with simplified chromosome name, gi12345)
        """
        print("starting depth test")
        depthdir = os.path.join(self.ref_dir, "samtools_depth_test_files")
        # ref = os.path.join(depthdir, "test_ref.fasta")
        test_bam = os.path.join(depthdir, "newref.bam")
        region = (1, 10000000)
        results = []
        for i in [region]:
            results.append(get_samtools_depths(samtools_exe=self.samtools_exe,
                                               bam=test_bam,
                                               chrom="gi12345", start=region[0],
                                               end=region[1],
                                               region=None,
                                               prep=True,
                                               logger=logger))
        print(results)
        self.assertEqual(round(results[0][1], 4), .9945)

    def test_check_fastqs_len_equal(self):
        failed = False
        try:
            check_fastqs_len_equal(self.ref_fasta, self.ref_fasta)
        except:
            failed = True
        self.assertFalse(failed)

    def test_check_fastqs_len_equal_fail(self):
        with self.assertRaises(ValueError):
            check_fastqs_len_equal(self.ref_fasta, self.ref_gb)

    def test_evaluate_spades(self):
        """ Test all the possible combination of return codes givena particular
        spades results
        """
        shutil.copyfile(self.good_contig,
                        os.path.join(self.test_dir,
                                     "contigs.fasta"))
        # assert os.path.exists(self.ref_fasta), "WTF"
        gen = SeedGenome(
            max_iterations=1,
            genbank_path=self.ref_gb,
            clustered_loci_txt=self.test_loci_file,
            output_root=self.test_dir,
            logger=logger)
        gen.loci_clusters = parse_clustered_loci_file(
            filepath=gen.clustered_loci_txt,
            gb_filepath=gen.genbank_path,
            output_root=self.test_dir,
            padding=100,
            circular=False,
            logger=logger)
        testmapping = LociMapping(
            name="test",
            iteration=1,
            assembly_subdir=self.test_dir,
            ref_fasta=self.ref_fasta,
            mapping_subdir=os.path.join(self.test_dir, "LociMapping"))
        testmapping2 = LociMapping(
            name="test",
            iteration=1,
            assembly_subdir_needed=False,
            assembly_subdir="nonexistantdir",
            ref_fasta=self.ref_fasta,
            mapping_subdir=os.path.join(self.test_dir, "LociMapping"))
        add_coords_to_clusters(seedGenome=gen, logger=logger)
        clu = gen.loci_clusters[0]

        code_0 = evaluate_spades_success(
            clu,
            mapping_ob=testmapping,
            proceed_to_target=False, target_len=None,
            min_assembly_len=5000,
            include_short_contigs=False, keep_best_contig=True,
            seqname='', logger=logger)
        self.assertEqual(code_0, 0)
        code_1 = evaluate_spades_success(
            clu,
            mapping_ob=testmapping,
            proceed_to_target=False, target_len=None,
            min_assembly_len=10000,
            include_short_contigs=True, keep_best_contig=True,
            seqname='', logger=logger)
        self.assertEqual(code_1, 1)
        code_1a = evaluate_spades_success(  # test proceed_to_target
            clu,
            mapping_ob=testmapping,
            proceed_to_target=True, target_len=6000,
            min_assembly_len=5000,
            include_short_contigs=True, keep_best_contig=True,
            seqname='', logger=logger)
        self.assertEqual(code_1a, 1)
        code_2 = evaluate_spades_success(
            clu,
            mapping_ob=testmapping,
            proceed_to_target=False, target_len=None,
            min_assembly_len=10000,
            include_short_contigs=False, keep_best_contig=True,
            seqname='', logger=logger)
        self.assertEqual(code_2, 2)
        code_3 = evaluate_spades_success(
            clu,
            mapping_ob=testmapping2,
            proceed_to_target=False, target_len=None,
            min_assembly_len=5000,
            include_short_contigs=False, keep_best_contig=True,
            seqname='', logger=logger)
        self.assertEqual(code_3, 3)

    def test_make_quast_table(self):
        paths = [os.path.join(self.ref_dir, "quast_1", "report.tsv"),
                 os.path.join(self.ref_dir, "quast_2", "report.tsv")]
        quast_dict = make_quick_quast_table(
            pathlist=paths, write=False, writedir=None, logger=logger)
        first10_printout = [
            "# N's per 100 kbp: ['0.00', '0.00']",
            "# contigs: ['2', '11']",
            "# contigs (>= 0 bp): ['3', '20']",
            "# contigs (>= 1000 bp): ['2', '10']",
            "# contigs (>= 10000 bp): ['2', '3']",
            "# contigs (>= 25000 bp): ['1', '0']",
            "# contigs (>= 5000 bp): ['2', '7']",
            "# contigs (>= 50000 bp): ['1', '0']",
            "# indels per 100 kbp: ['3.98', '0.00']",
            "# local misassemblies: ['0', '0']",
            "# misassembled contigs: ['0', '0']"]
        counter = 0
        for k, v in sorted(quast_dict.items()):
            self.assertEqual(first10_printout[counter],
                             "{0}: {1}".format(k, v))
            if counter == 10:
                break
            else:
                counter = counter + 1

    def test_partition_mapping(self):
        # this mostly does system calls; cant really test smoothlu
        pass

    @unittest.skipIf(shutil.which("smalt") is None, "smalt executable not found, skipping."+
                     "If this isnt an error from travis deployment, you probably "+
                     "should install it")
    def test_prepare_next_mapping(self):
        gen = SeedGenome(
            max_iterations=1,
            genbank_path=self.ref_gb,
            clustered_loci_txt=self.test_loci_file,
            output_root=self.test_dir,
            logger=logger)
        testmapping = LociMapping(
            name="test",
            iteration=1,
            assembly_subdir=self.test_dir,
            ref_fasta=self.ref_fasta,
            mapping_subdir=os.path.join(self.test_dir, "LociMapping"))
        testngs = NgsLib(
            name="test",
            master=True,
            make_dist=True,
            readF=self.ref_Ffastq,
            readR=self.ref_Rfastq,
            ref_fasta=self.ref_fasta,
            mapper_exe=self.smalt_exe,
            logger=logger)
        gen.master_ngs_ob = testngs
        gen.loci_clusters = parse_clustered_loci_file(
            filepath=gen.clustered_loci_txt,
            gb_filepath=gen.genbank_path,
            output_root=self.test_dir,
            padding=1000,
            circular=False,
            logger=logger)
        add_coords_to_clusters(seedGenome=gen,
                               logger=logger)
        gen.next_reference_path = gen.ref_fasta
        #
        for cluster in gen.loci_clusters:
            cluster.master_ngs_ob = gen.master_ngs_ob

        map_to_genome_ref_smalt(mapping_ob=testmapping, ngsLib=testngs,
                                cores=4, samtools_exe=self.samtools_exe,
                                smalt_exe=self.smalt_exe, score_minimum=48,
                                scoring="match=1,subst=-4,gapopen=-4,gapext=-3",
                                step=3, k=5, logger=logger)
        clu = gen.loci_clusters[0]
        self.assertEqual(len(clu.mappings), 0)
        prepare_next_mapping(
            cluster=clu, seedGenome=gen, samtools_exe=self.samtools_exe,
            flank=0,
            logger=logger)
        new_name = "NC_011751.1_cluster_{0}".format(clu.index)
        self.assertEqual(clu.mappings[-1].name, new_name)

    def test_mapped_partition_cmds(self):
        gen = SeedGenome(
            max_iterations=1,
            genbank_path=self.ref_gb,
            clustered_loci_txt=self.test_loci_file,
            output_root=self.test_dir,
            logger=logger)
        testmapping = LociMapping(
            name="test",
            iteration=1,
            assembly_subdir=self.test_dir,
            ref_fasta=self.ref_fasta,
            mapping_subdir=os.path.join(self.test_dir, "LociMapping"))
        gen.loci_clusters = parse_clustered_loci_file(
            filepath=gen.clustered_loci_txt,
            gb_filepath=gen.genbank_path,
            output_root=self.test_dir,
            padding=1000,
            circular=False,
            logger=logger)
        add_coords_to_clusters(seedGenome=gen, logger=logger)
        gen.loci_clusters[0].global_start_coord = 5000
        gen.loci_clusters[0].global_end_coord = 10000
        # print(gen.loci_clusters[0].__dict__)
        cmds, region = make_mapped_partition_cmds(
            cluster=gen.loci_clusters[0],
            mapping_ob=testmapping,
            seedGenome=gen,
            samtools_exe=self.samtools_exe,
            logger=logger)
        ref_cmds = [
            '{0} sort {1} > {2}'.format(
                self.samtools_exe,
                os.path.join(
                    self.test_dir,
                    "NC_011751.1_mapping_for_iter_0/" +
                    "NC_011751.1_mapping_for_iter_0_iteration_0.bam"),
                os.path.join(
                    self.test_dir,
                    "NC_011751.1_mapping_for_iter_0/" +
                    "NC_011751.1_mapping_for_iter_0_iteration_0_sorted.bam")),
            "{0} index {1}".format(
                self.samtools_exe,
                os.path.join(
                    self.test_dir,
                    "NC_011751.1_mapping_for_iter_0/" +
                    "NC_011751.1_mapping_for_iter_0_iteration_0_sorted.bam")),
            '{0} view -o {1} {2} NC_011751.1:5000-10000'.format(
                self.samtools_exe,
                os.path.join(self.test_dir,
                             "LociMapping/test_iteration_1.bam"),
                os.path.join(
                    self.test_dir,
                    "NC_011751.1_mapping_for_iter_0/" +
                    "NC_011751.1_mapping_for_iter_0_iteration_0_sorted.bam"))]
        # print(region)
        self.assertEqual(region, "NC_011751.1:5000-10000")
        for i, cmd in enumerate(cmds):
            self.assertEqual(ref_cmds[i], cmd)

    def test_unmapped_partition_cmds(self):
        gen = SeedGenome(
            max_iterations=1,
            genbank_path=self.ref_gb,
            clustered_loci_txt=self.test_loci_file,
            output_root=self.test_dir,
            logger=logger)

        unmapped_cmds = make_unmapped_partition_cmds(
            mapped_regions=["test:1-6", "test:3-77"],
            samtools_exe=self.samtools_exe, seedGenome=gen)
        mapped_txt = os.path.join(
            self.test_dir,
            "NC_011751.1_mapping_for_iter_0/" +
            "NC_011751.1_mapping_for_iter_0_iteration_0_mapped.txt")

        ref_unmapped_cmds = [
            "{0} view -o {1} -h {2}".format(
                self.samtools_exe,
                os.path.join(self.test_dir,
                             "NC_011751.1_mapping_for_iter_0/" +
                             "NC_011751.1_mapping_for_iter_0_iteration_0.sam"),
                os.path.join(self.test_dir,
                             "NC_011751.1_mapping_for_iter_0/" +
                             "NC_011751.1_mapping_for_iter_0_iteration_0.bam")),
            "{0} view {1} test:1-6 | cut -f1 >> {2}".format(
                self.samtools_exe,
                os.path.join(self.test_dir,
                             "NC_011751.1_mapping_for_iter_0/" +
                             "NC_011751.1_mapping_for_iter_0_iteration_0_sorted.bam"),
                mapped_txt),
            "{0} view {1} test:3-77 | cut -f1 >> {2}".format(
                self.samtools_exe,
                os.path.join(self.test_dir,
                             "NC_011751.1_mapping_for_iter_0/" +
                             "NC_011751.1_mapping_for_iter_0_iteration_0_sorted.bam"),
                mapped_txt),
            "sort -u {0}".format(mapped_txt),
            "LC_ALL=C grep -w -v -F -f {0} < {1} > {2}".format(
                mapped_txt,
                os.path.join(self.test_dir,
                             "NC_011751.1_mapping_for_iter_0/" +
                             "NC_011751.1_mapping_for_iter_0_iteration_0.sam"),
                os.path.join(self.test_dir,
                             "NC_011751.1_mapping_for_iter_0/" +
                             "NC_011751.1_mapping_for_iter_0_iteration_0_unmapped.sam"))
        ]
        for i, ref in enumerate(ref_unmapped_cmds):
            self.assertEqual(unmapped_cmds[i], ref)
        # self.to_be_removed.append(mapped_txt)

    def test_make_faux_genome(self):
        gen = SeedGenome(
            max_iterations=1,
            genbank_path=self.ref_gb,
            clustered_loci_txt=self.test_loci_file,
            output_root=self.test_dir,
            logger=logger)
        gen.loci_clusters = parse_clustered_loci_file(
            filepath=gen.clustered_loci_txt,
            gb_filepath=gen.genbank_path,
            output_root=self.test_dir,
            padding=100,
            circular=False,
            logger=logger)
        add_coords_to_clusters(seedGenome=gen, logger=logger)

        for clu in gen.loci_clusters:
            clu.keep_contig = True
            clu.mappings.append(LociMapping(
                name="test",
                iteration=1,
                assembly_subdir=self.test_dir,
                ref_fasta=self.ref_fasta,
                assembled_contig=self.ref_fasta,
                mapping_subdir=os.path.join(
                    self.test_dir, "LociMapping_for_text_faux_genome")))

        # map_to_genome_ref_smalt(
        #     mapping_ob=gen.iter_mapping_list[0],
        #     ngsLib=gen.master_ngs_ob,
        #     cores=4,
        #     samtools_exe=self.samtools_exe,
        #     smalt_exe=self.smalt_exe,
        #     score_minimum=None,
        #     step=3, k=5,
        #     scoring="match=1,subst=-4,gapopen=-4,gapext=-3",
        #     logger=logger)
        # partition_mapping(seedGenome=gen,
        #                   logger=logger,
        #                   samtools_exe=self.samtools_exe,
        #                   flank="50:50",
        #                   cluster_list=gen.loci_clusters)

        make_faux_genome(cluster_list=gen.loci_clusters, seedGenome=gen,
                         iteration=1, output_root=self.test_dir, nbuff=10000,
                         logger=logger)

    def test_get_final_assemblies_cmds(self):
        gen = SeedGenome(
            max_iterations=1,
            genbank_path=self.ref_gb,
            clustered_loci_txt=self.test_loci_file,
            output_root=self.test_dir,
            assembled_seeds="tralalalala",
            logger=logger,
            master_ngs_ob=NgsLib(
                name="test",
                master=False,
                readF=self.ref_Ffastq,
                readR=self.ref_Rfastq,
                ref_fasta=self.ref_fasta,
                mapper_exe=self.smalt_exe))
        test_exes = Exes(samtools=self.samtools_exe,
                         quast=self.quast_exe,
                         python2_7=self.python2_7_exe,
                         smalt=self.smalt_exe,
                         spades=self.spades_exe,
                         bwa=self.bwa_exe,
                         check=False,
                         method="bwa")
        # I want this  generalized, so replace the acual path with spades.py
        test_exes.spades = "spades.py"
        test_exes.quast = "quast.py"
        test_exes.python2_7 = "python2"
        gen.ref_fasta = self.ref_fasta
        final_cmds, quast_reports = \
            get_final_assemblies_cmds(
                seedGenome=gen, exes=test_exes,
                skip_control=False, kmers="33,77,99", logger=logger)
        final_spades_cmds_ref = [
            str(
                "if [ -s {1} ] && [ -s {2} ] ; then " +
                "{0} --careful -k 33,77,99 --pe1-1 {1} " +
                "--pe1-2 {2} --trusted-contigs {3}  -o {4} " +
                "; else echo 'input lib not found, " +
                "skipping this SPAdes call' ; fi"
            ).format(
                self.spades_exe, self.ref_Ffastq, self.ref_Rfastq,
                gen.assembled_seeds,
                os.path.join(self.test_dir, "final_de_fere_novo_assembly")),
            str(
                "if [ -s {1} ] && [ -s {2} ] ; then " +
                "{0} --careful -k 33,77,99 --pe1-1 {1} " +
                "--pe1-2 {2}   -o {3} " +
                "; else echo 'input lib not found, " +
                "skipping this SPAdes call' ; fi"
            ).format(
                self.spades_exe, self.ref_Ffastq, self.ref_Rfastq,
                os.path.join(self.test_dir, "final_de_novo_assembly"))]
        final_quast_cmds_ref = [
            str(
                '{0} {1} -R {2} {3} -o {4}'
            ).format(
                self.python2_7_exe, self.quast_exe,
                self.ref_fasta,
                os.path.join(self.test_dir, "final_de_fere_novo_assembly",
                             "contigs.fasta"),
                os.path.join(self.test_dir, "quast_de_fere_novo")),
            str(
                '{0} {1} -R {2} {3} -o {4}'
            ).format(
                self.python2_7_exe, self.quast_exe,
                self.ref_fasta,
                os.path.join(self.test_dir, "final_de_novo_assembly",
                             "contigs.fasta"),
                os.path.join(self.test_dir, "quast_de_novo"))]
        for i, cmd in enumerate([x[0] for x in final_cmds]):
            self.assertEqual(final_spades_cmds_ref[i], cmd)
        for i, cmd in enumerate([x[1] for x in final_cmds]):
            self.assertEqual(final_quast_cmds_ref[i], cmd)

    def tearDown(self):
        """ delete temp files if no errors
        """
        for filename in self.to_be_removed:
            os.unlink(filename)
        pass

if __name__ == '__main__':
    unittest.main()
    # tools_needed = ["samtools", "smalt", "spades.py"]
    # for i in tools_needed:
    #     if not check_installed_tools(i, hard=False):
    #         print("Error! Can't run without install of {0} in PATH".format(i))
    #         sys.exit(1)
    # logger = logging



"""
Setup for nrw_python_utils

A setuptools based setup module.
See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
import re
from codecs import open
from os import path
import sys
from pip.req import parse_requirements

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
# with open(path.join(here, 'README.md'), encoding='utf-8') as f:
#     long_description = f.read()

# from pyani
with open(path.join(here, 'riboSeed', '__init__.py'), 'r') as fh:
    for line in fh:
        m = re.search(r"^__version__ = '(?P<version>[^']+)'$", line)
        if m:
            init_version = m.group('version')
            break

if sys.version_info <= (3, 0):
    sys.stderr.write("ERROR: riboseed requires Python 3.5 " +
                     "or above...exiting.\n")
    sys.exit(1)

## parse requirements file
install_reqs = parse_requirements(path.join(here, "requirements.txt"),
                                  session=False)
requirements = [str(ir.req) for ir in install_reqs]
setup(
    name='riboSeed',

    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version=init_version,

    description='riboSeed: assemble across rDNA regions',
    # long_description=long_description,
    long_description="""
    its a blast, you gotta try it!  check out the GitHub
    repo for the read README.md file
    """,

    # The project's main homepage.
    url='https://github.com/nickp60/riboSeed',

    # Author details
    author='Nick Waters',
    author_email='nickp60@gmail.com',

    # Choose your license
    license='MIT',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: MIT License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        #'Programming Language :: Python :: 2',
        #'Programming Language :: Python :: 2.6',
        #'Programming Language :: Python :: 2.7',
        #'Programming Language :: Python :: 3',
        #'Programming Language :: Python :: 3.3',
        #'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],

    # What does your project relate to?
    keywords='sample bioinformatics setuptools development',

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    # packages=['riboSeed'],
    packages=find_packages(),

    # Alternatively, if you want to distribute just a my_module.py, uncomment
    # this:
    #   py_modules=["my_module"],

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=requirements,
    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    #extras_require={
    #    'dev': ['check-manifest'],
    #    'test': ['coverage'],
    #},

    # If there are data files included in your packages that need to be
    # installed, specify them here.  If using Python 2.6 or less, then these
    # have to be included in MANIFEST.in as well.
    #package_data={
    #    'sample': ['package_data.dat'],
    #},

    # Although 'package_data' is the preferred approach, in some case you may
    # need to place data files outside of your packages. See:
    # http://docs.python.org/3.4/distutils/setupscript.html#installing-additional-files # noqa
    # In this case, 'data_file' will be installed into '<sys.prefix>/my_data'
    #data_files=[('my_data', ['data/data_file'])],

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    #entry_points={
    #    'console_scripts': [
    #        'sample=sample:main',
    #    ],
    #},
    # Using the scripts keyword here for speed, as the entry points are not
    # yet defined
    scripts=['riboSeed/riboSelect.py', 'riboSeed/riboSeed.py',
             'riboSeed/riboSnag.py'],
)



#!/usr/bin/env python3.5
"""
Input:
- genbank file
- location:  rRNA (or RRNA, or something similar)

Output:
-text file with sequence name and colon-separated locus_tags,
  one for each region
NC33352.5  ECUMN_23S_3:ECUMN_16S_3:ECUMN_5S_4


USAGE:
 $ python3.5 riboSelect.py genome.gbk --output /path/to/output
"""
import os
import csv
import subprocess
import datetime
import argparse
import sys
import time
from Bio import SeqIO

from pyutilsnrw.utils3_5 import set_up_logging, multisplit


def get_args():  # pragma: no cover
    """get the arguments as a main parser with subparsers
    for named required arguments and optional arguments
    """
    parser = argparse.ArgumentParser(description="This is used to extract" +
                                     " rRNA regions from a gb file, returns" +
                                     "a text file with the clusters")
    parser.add_argument("genbank_genome", help="Genbank file (WITH SEQUENCE)")
    requiredNamed = parser.add_argument_group('required named arguments')
    requiredNamed.add_argument("-o", "--output",
                               help="output directory;"
                               "default: %(default)s",
                               default=os.getcwd(),
                               type=str, dest="output")
    optional = parser.add_argument_group('optional arguments')
    optional.add_argument("-f", "--feature",
                          help="Feature, rRNA or RRNA; default: %(default)s",
                          default='rRNA', type=str)
    optional.add_argument("-s", "--specific_features",
                          help="colon:separated -- specific features"
                          "; default: %(default)s",
                          default='16S:23S:5S', type=str)
    optional.add_argument("--keep_temps",
                          help="view intermediate clustering files"
                          "default: %(default)s",
                          action='store_true',
                          default=False, dest="keep_temps")
    optional.add_argument("--clobber",
                          help="overwrite previous output files: "
                          "default: %(default)s", action='store_true',
                          default=False, dest="clobber")
    # TODO Implement this for work with really fragmented genomes
    # optional.add_argument("--nocluster",
    #                       help="do not bother clustering; treat all " +
    #                       "occurances on a sequence as a cluster" +\
    #                       "default: %(default)s", action='store_true',
    #                       default=False, dest="nocluster")
    optional.add_argument("-c", "--clusters",
                          help="number of rDNA clusters;"
                          "if submitting multiple records, must be a "
                          "colon:separated list that matches number "
                          "of genbank records.  Default is inferred from "
                          "specific feature with fewest hits", default='',
                          type=str, dest="clusters")
    optional.add_argument("-v", "--verbosity",
                          dest='verbosity', action="store",
                          default=2, type=int,
                          help="1 = debug(), 2 = info(), 3 = warning(), "
                          "4 = error() and 5 = critical(); "
                          "default: %(default)s")
    optional.add_argument("--debug", dest="debug", action="store_true",
                          help="Enable debug messages", default=False)
    args = parser.parse_args()
    return args


def count_feature_hits(all_feature, genome_seq_records,
                       specific_features, locus_tag_dict):
    """ given genome seq records, specific_features, and a
    locus_tag_dict from get_filtered_locus_tag_dict, return two structures:
    - nfeatures_occur {record.id, [[16s, 5],[23s, 4],[5s,6]]}
    - nfeatures_simple {record.id [record.id, [5,4,6]}
    """
    #  This bit counts the number of hits per specific feature.
    if not all_feature:
        nfeatures_occur = {}  # makes  {genome : [['18S', 5],['28S',3]]}
        nfeat_simple = {}  # makes  {genome : [5, 3]}
        for record in genome_seq_records:
            hit_list = []  # [specific feature, count] list
            hit_list_simple = []  # [count] list
            for i in specific_features:
                hits = 0
                subset = {k: v for k, v in locus_tag_dict.items()
                          if record.id in v}
                for k, v in subset.items():
                    # hint: v[-1] should be the product annotation
                    if any([i in x for x in v[-1]]):
                        hits = hits + 1
                    else:
                        pass
                hit_list.append([i, hits])
                hit_list_simple.append(hits)
            nfeatures_occur[record.id] = (hit_list)
            nfeat_simple[record.id] = hit_list_simple
    else:
        nfeatures_occur, nfeat_simple = None, None
    return(nfeatures_occur, nfeat_simple)


def get_filtered_locus_tag_dict(genome_seq_records, feature="rRNA",
                                specific_features="16S:23S",
                                verbose=True, logger=None):
    """ Given a LIST (as of 20160927) or genbank records,
    returns dictionary of index:locus_tag id pairs for all
    "feature"  entries.  This then gets clustered.
    requires having locus tag in your genbank file.  Non-negitable.
    should be prokka-friendly, so if you have a gb file with legacy
    annotations, just run through prokka (with an rRNA caller)
    20160922 This was changed to add checking for rRNA type from
    ribosome product annoation, not ht locus tag.
    if specific features is None, return all with type == feature
    Changed from having index used for clustering to using the first coord
    """
    assert isinstance(genome_seq_records, list), \
        'must pass list of genomes to function, even if single genome '
    # if specific_features is None:
    #     all_feature = True
    # else:
    #     all_feature = False
    if specific_features is not None:
        specific_features = specific_features.split(":")
    locus_tag_dict = {}  # recipient structure
    # loop through records
    for record in genome_seq_records:
        loc_number = 0  # counter that for index of hits; this is clustered
        logger.debug("scanning %s", record.id)
        for feat in record.features:
            if feat.type in feature:
                try:
                    locustag = feat.qualifiers.get("locus_tag")[0]
                except TypeError:
                    if verbose:
                        logger.debug("no locus tag for this feature!")
                    continue
                product_list = multisplit([",", " ", "-", "_"],
                                          feat.qualifiers.get("product")[0])
                coords = [feat.location.start.position + 1,
                          feat.location.end.position]
                logger.debug(product_list)
                logger.debug(coords)
                # if either specific feature is found in product or
                # only interested in all features, add locus to dict
                if (specific_features is None or
                    (specific_features is not None and
                     any([x in specific_features for x in product_list]))):
                    # key is start coord
                    locus_tag_dict[(record.id, coords[0])] = [loc_number,
                                                              record.id,
                                                              locustag,
                                                              feat.type,
                                                              product_list]
                else:
                    logger.debug("Not adding this feat to " +
                                 "list: %s", product_list)
                    pass
                loc_number = loc_number + 1  # increment index after feature
            else:
                if verbose:
                    logger.debug("skipping: %s", feat.type)
                loc_number = loc_number + 1  # increment index after feature
        # this is a soft warning, as we want to be able to loop
        # through all records before worrying
        if len(locus_tag_dict) < 1 and logger:
            logger.info(str("no locus tags found in {0} for " +
                            "{1} features with annotated products matching " +
                            "{2}!\n").format(record.id,
                                             feature, specific_features))
    locus_tag_dict = locus_tag_dict
    if verbose and logger:
        for key in sorted(locus_tag_dict):
            logger.debug("%s: %s;", key, locus_tag_dict[key])

    # count the occuraces of each feature per genbank record
    nfeatures_occur, \
        nfeat_simple = count_feature_hits(
            all_feature=specific_features is None,
            genome_seq_records=genome_seq_records,
            specific_features=specific_features,
            locus_tag_dict=locus_tag_dict)
    return(locus_tag_dict, nfeatures_occur, nfeat_simple)


def pure_python_kmeans(data, centers=3, kind=int, DEBUG=True):
    """giveb 1d list of numberic data and number of centers, returns a
    csv with the data and cluster, and LP's disapointment
    """
    # handle single instance "clusters" when data is same or less than centers
    if len(data) <= centers:
        # {"index": [coord]}
        return {str(i): [j] for i, j in enumerate(data)}

    with open(os.path.join(os.getcwd(),
                           "pure_python_kmeans_list.csv"), "w") as f:
        for i in data:
            f.write("".join([str(i), "\n"]))
    rcmds = ["# Generated by riboSelect.py on {0}".format(time.asctime()),
             "centers <- {0}".format(centers),
             str("data <- read.csv('pure_python_kmeans_list.csv', " +
                 "header=F, col.names=c('index'))"),
             "set.seed(27)",
             "km <- kmeans(data[,1], nstart=100,iter.max=100,centers=centers)",
             "data[,2] <- km[1]",
             "data <- data[order(data[,1]),  ]",
             str("write.table(data, 'pure_python_kmeans_list.csv', " +
                 "sep=',', row.names=F, col.names=F)")
             ]
    with open(os.path.join(os.getcwd(), "km_script.R"), "w") as f:
        for i in rcmds:
            f.write(i + "\n")
    subprocess.run("Rscript km_script.R", shell=sys.platform != 'win32',
                   check=True)
    with open("pure_python_kmeans_list.csv", mode='r') as infile:
        reader = csv.reader(infile)
        indexClusterDict = {}
        for row in reader:
            try:
                if row[1] in indexClusterDict:
                    indexClusterDict[row[1]].append(row[0])
                else:
                    indexClusterDict[row[1]] = [row[0]]
            except:
                raise ImportError("error constructing dictionary from csv; "
                                  "possibly due to type casting? adjust the "
                                  " 'kind' arg to string if in doubt")
    if not DEBUG:
        os.remove(os.path.join(os.getcwd(), "pure_python_kmeans_list.csv"))
        os.remove(os.path.join(os.getcwd(), "km_script.R"))
    cast_dict = {}
    if kind not in [int, str, float]:
        raise TypeError("can only attempt coersion to int, str, and float; " +
                        "cant coerce to {0}.".format(kind))
    for k, v in indexClusterDict.items():
        cast_dict[k] = [kind(x) for x in v]
    return cast_dict


if __name__ == "__main__":
    args = get_args()
    output_root = os.path.abspath(os.path.expanduser(args.output))
    # Create output directory only if it does not exist
    try:
        os.makedirs(args.output)
    except FileExistsError:
         # leaving comment char'#' added for stream output usage
        print("#Selected output directory %s exists" %
              args.output)
        if not args.clobber:
            print("exiting")
            sys.exit(1)
        else:
            print("# continuing, and risking potential loss of data")
    logger = set_up_logging(
        verbosity=args.verbosity,
        outfile=str("%s_riboSelect_log.txt" %
                    os.path.join(output_root,
                                 time.strftime("%Y%m%d%H%M"))),
        name=__name__)

    # log = sys.stderr.write  # to keep streaming clean if this goes that route
    logger.info("Current usage:\n{0}\n".format(" ".join(sys.argv[1:])))
    logger.debug("All settings used:")
    for k, v in sorted(vars(args).items()):
        logger.debug("%s: %s", k, v)
    date = str(datetime.datetime.now().strftime('%Y%m%d'))

    # Check if output file exists; if so, remove it
    output_path = os.path.join(args.output,
                               str("riboSelect_grouped_loci.txt"))
    if os.path.exists(output_path):
        if args.clobber:
            logger.info("removing existing output file\n")
            os.remove(output_path)
        else:
            logger.error("Existing output file found!")
            sys.exit(1)

    # get genome records into a list
    with open(args.genbank_genome) as fh:
        genome_records = list(SeqIO.parse(fh, 'genbank'))
    print("Loaded %d records" % len(genome_records))

    # get list of loci matching feature and optionally specific features
    # also returns nfeat, a dict of feature count by genbank id
    logger.debug(
        str(
            "searching {0} for {1} features containing {2} in the " +
            "product annotation").format(
            str([x.id for x in genome_records]),
            args.feature,
            str([x for x in args.specific_features.split(":")])))
    lociDict, nfeat, nfeat_simple = \
        get_filtered_locus_tag_dict(genome_seq_records=genome_records,
                                    feature=args.feature,
                                    specific_features=args.specific_features,
                                    verbose=args.debug,
                                    logger=logger)

    # default case, clusters are inferred
    # if not, must be equal to the length of genbank records
    if args.clusters != "":
        try:
            centers = [int(x) for x in args.clusters.split(":")]
            logger.info(str(centers))
        except:
            logger.error("cannot coerce --clusters to integer after " +
                         "splitting on colons!\n")
            sys.exit(1)
    else:
        centers = [0 for x in genome_records]

    # if unequal lengths, throw error
    # logger.info clusters for accession for user to verify
    if len(genome_records) != len(centers):
        logger.error("centers must be the same length as number" +
                     " of genbank records!\n")
        sys.exit(1)

    #####
    ##### for each genbank record, process, and append any hits to outfile
    #####
    logger.debug("All loci:")
    for k, v in lociDict.items():
        logger.debug(str(k) + "\t" + str(v))
    for i in range(0, len(genome_records)):

        logger.info("Processing {0}\n".format(genome_records[i].id))
        # if user gives clusters, make sure it matches the length:
        if args.clusters:
            logger.info("using %s clusters for %s\n",
                        centers[i], genome_records[i].id)
        # get subset of lociDict for that id
        subset = {key: value for key, value in lociDict.items() if
                  genome_records[i].id in value}
        # skip if that doesnt have any hits
        logger.debug("Subset loci:")
        for k, v in subset.items():
            logger.debug(str(k) + "\t" + str(v))

        if len(subset) == 0:
            logger.info("no hits in {0}\n".format(genome_records[i].id))
            continue
        logger.info("hits in {0}\n".format(genome_records[i].id))

        #  find nfeat for this genbank id by subsetting;
        # is this a bad way of doesnt things?
        logger.debug("centers: {0}".format(centers))
        logger.debug("nfeat_simple: {0}".format(nfeat_simple))
        if nfeat_simple is None and centers[i] == 0:
            logger.error("No specific features submitted, cannot calculate " +
                         " number centers needed for clustering.  Please" +
                         "submit the desired number of clusters with the  " +
                         "--clusters argument!\n")
            sys.exit(1)
        rec_nfeat = list({k: v for k, v in nfeat_simple.items() if
                          genome_records[i].id in k}.values())[0]
        logger.debug("rec_nfeat: {0}".format(rec_nfeat))
        indexes = [x[1] for x in list(subset)]  # get index back from tuple key
        ## if centers[i] is 0, try max and min sequentially; if that fails skip
        if centers[i] == 0:
            if min(rec_nfeat) == 0:
                current_centers = max(rec_nfeat)
            else:
                current_centers = min(rec_nfeat)
            if current_centers == 0:
                logger.info("skipping the clustering for {0}\n".format(i))
                continue
            # indexClusters = pure_python_kmeans(subset.keys(),
            indexClusters = pure_python_kmeans(indexes,
                                               centers=current_centers,
                                               DEBUG=args.keep_temps)
        else:
            ## if centers[i] is not 0, use it
            current_centers = centers[i]
        # Perform actual clustering
        try:
            # indexClusters should be like { "1": [3,4,6], "2": [66,45,63]}
            indexClusters = pure_python_kmeans(indexes,
                                               centers=current_centers,
                                               DEBUG=args.keep_temps)
        except Exception as e:
            logger.error(e)
            sys.exit(1)

        with open(output_path, "a") as outfile:
            outfile.write("# Generated cluters for {0} on {1}\n".format(
                i, date))
            outfile.write("#$ FEATURE {0}\n".format(args.feature))
            sys.stdout.write("#$ FEATURE {0}\n".format(args.feature))
            sys.stdout.write("# Generated cluters for {0} on {1}\n".format(
                i, date))
            for k, v in indexClusters.items():
                # for each k:v, this replaces the index in v with the locus tag
                # from subset, and writes it out in the way that plays nice
                # riboSeed
                outstr = str(genome_records[i].id + " " +
                             str(":".join([subset[(genome_records[i].id,
                                                   int(x))][2] for x in v])) +
                             '\n')
                outfile.write(outstr)
                # this should be the only thing going to stdout.
                sys.stdout.write(outstr)



#!/usr/bin/env python
"""
With vesion 0.8.0 of riboseed, we moved away from the flexibility of getting any region from a genome.  This is a stand in to create a list of other loci of interest. 

version 0.8.1
Minor version changes:
 - This is the first instance of otherSelect.py and riboSelect.py.
#TODO:
- Make this less awful
- No seriously, this is awful
Input:
- genbank file
- location: cds, gene, or rRNA
- specific features : 16S, 5S
- specific regions : upstream, downstream, etc
- region lengths: int

Output:
-DNA fasta

USAGE:
 $ python riboSnag.py genome.gbk --features rRNA --specific_features 16S:23S --feature_regs upstream:downstream --feature_regs_len 200:400
     --within_feature_length 50  --output /path/to/output
"""
import re
import os
import pprint
#import random
#import sys
import csv
import subprocess
import datetime
import argparse
import sys
from Bio import SeqIO
from Bio.SeqRecord import SeqRecord
from Bio.Alphabet import IUPAC
from Bio.Seq import Seq
pp = pprint.PrettyPrinter(indent=4)

#%%


def get_args(DEBUG=False):
    # dummy_args = argparse.Namespace(input_genome_path="NC_011751.1.gb",
    #                                 feature="rRNA",
    #                                 pre_specific_features="16S:23S",
    #                                 pre_feature_regs="upstream:downstream",
    #                                 pre_feature_regs_len="200:400",
    #                                 within=50,
    #                                 output=os.path.expanduser("~/Desktop/"))
    # if DEBUG:  # ie, if running interactively
    #     print("Using Dummy Arguments")
    #     return(dummy_args)
    parser = argparse.ArgumentParser(description="This is used to extract regions of interest based\
                                                  on coordinates and feature types.")
    parser.add_argument("genbank_genome", help="Genbank file (WITH SEQUENCE)")
    parser.add_argument("-f", "--feature", help="Feature, such as CDS,tRNA, rRNA; default: %(default)s",
                        default='rRNA', type=str)
    parser.add_argument("-s", "--specific_features", help="colon:separated -- specific features\
                         to be grepped from product, such as 16S or tRNA-Ala; default: %(default)s",
                        default='16S:23S', type=str)
    parser.add_argument("-r", "--feature_regions", help="colon:separated -- upstream, \
                         upstream_within, downstream, downstream_within, identity, identityN; default: %(default)s",
                        default="upstream_within:downstream_within", type=str)
    parser.add_argument("-l", "--feature_regions_lengths", help="colon:separated -- length of\
                         region; default: %(default)s",  default="100:100", type=str)
    parser.add_argument("-w", "--within_feature_length", help="if using up/downstream_within, \
                         where you return some of the feature itself, this dictates how many bp; default: %(default)s",
                        default=50, type=int)
    parser.add_argument("-o", "--output", help="output directory; default: %(default)s", default=os.getcwd(),
                        type=str, dest="output")
    # parser.add_argument("--DEBUG", help="run on test data; default: %(default)s", default=False,
    #                     action="store_true", dest="DEBUG")
    args = parser.parse_args()
    # if args.DEBUG:  # ie, if running debug from cline
    #     print("Using Dummy Arguments")
    #     return(dummy_args)
    # else:
    return(args)


def check_single_scaffold(input_genome_path):
    """Test for single scaffold
    """
    print("testing for multiple scaffolds...")
    counter = -1  # if all goes well, returns 0, else returns 1 or more or -1
    for line in open(input_genome_path, "r"):
        if re.search("ORIGIN", line) is not None:
            counter = counter + 1
    return(counter)


def get_genbank_record(input_genome_path):
    """reads the FIRST record only from a genbank file; will probably only work for first scaffold
    """
    print("Reading genbank file...")
    with open(input_genome_path) as input_genome_handle:
        genome_seq_record = next(SeqIO.parse(input_genome_handle, "genbank"))
    if genome_sequence[0: 100] == str("N" * 100):
        print("Careful: the first 100 nucleotides are N's; did you download the full .gb file?")
    return(genome_seq_record)


def get_genbank_seq(input_genome_path):
    """get the sequence from the FIRST record only in a genbank file
    """
    print("fetching nucleotide sequence from genbank file...")
    with open(input_genome_path) as input_genome_handle:
        genome_seq_record = next(SeqIO.parse(input_genome_handle, "genbank"))
    return(genome_seq_record.seq)


# def extract_coords(genome_seq_record, feature="rRNA", verbose=True):
#     """the magic (part 1) of this whole thing
#     """
#     loc_number = 0  # index for hits
#     loc_list = []  # recipient structure
#     for feat in genome_seq_record.features:
#         if feat.type != feature:  # skip if not features of interest
#             pass
#         else:
#             #  SeqIO subtracts one from start coord to make it pythonic. the +1 below undoes that
#             coords = [feat.location.start.position + 1, feat.location.end.position]
#             strand = feat.strand
#             product = feat.qualifiers.get("product")
#             if "locus_tag" in feat.qualifiers:
#                 locus_tag = feat.qualifiers.get("locus_tag")
#             else:
#                 locus_tag = "none"
#             loc_list.append([loc_number, coords, strand, product, locus_tag])
#             loc_number = loc_number + 1
#     if not loc_number > 0:
#         raise ValueError("no hits found!")
#         sys.exit(1)
#     if verbose:
#         print("Here are the detected region, strand, product, locus tag, \
#                and subfeatures of the results:") 
#         pp.pprint(loc_list)
#     return(loc_list)


def extract_coords_from_locus(genome_seq_record, locus_tag_list=[], verbose=True):
    """given a list of locus_tags, return a list of loc_number,coords, strand, product
    """
    loc_number = 0  # index for hits
    loc_list = []  # recipient structure
    for feat in genome_seq_record.features:
        try:
            if (feat.qualifiers.get("locus_tag")[0] in locus_tag_list) and\
               (feat.type == args.feature):
                #  SeqIO makes coords 0-based; the +1 below undoes that
                coords = [feat.location.start.position + 1, 
                          feat.location.end.position]
                strand = feat.strand
                product = feat.qualifiers.get("product")
                locus_tag = feat.qualifiers.get("locus_tag")
                loc_list.append([loc_number, coords, strand, 
                                 product, locus_tag])
                loc_number = loc_number + 1
            else:
                pass
        except:
            pass
    if not loc_number > 0:
        raise ValueError("no hits found!")
        sys.exit(1)
    if verbose:
        print("Here are the detected region, strand, product, locus tag, \
               and subfeatures of the results:") 
        pp.pprint(loc_list)
    return(loc_list)

def pure_python_kmeans(data, centers=3):
    """giveb 1d data and number of centers, returns a
    csv with the data and cluster, and LP's disapointment
    """
    with open(os.path.join(os.getcwd(), "list.csv"), "w") as f:
        for i in data:
            f.write("".join([str(i), "\n"]))
    
    with open(os.path.join(os.getcwd(), "km_script.R"), "w") as f:
        rcmds= ["centers<-{0}".format(centers),
                "data<-read.csv('list.csv', header=F, col.names=c('index'))", 
                "km <- kmeans(data[,1], nstart=100, iter.max=100, centers=centers)",
                "data[,2]<-km[1]",
                "data<- data[order(data[,1]),  ]",
                "write.csv(data, 'list.csv',  row.names=F)"]
        for i in rcmds:
                f.write(i)
                f.write('\n')
    subprocess.call("Rscript km_script.R", shell=sys.platform!='win32')

def get_filtered_locus_tag_dict(genome_seq_record, feature="rRNA", verbose=True):
    """returns dictionary of index:locus_tag id pairs for all 
    "feature"  entries.  This then gets clustered
    requires having locus tag in your genbank file.  Non-negitable
    """
    loc_number = 0 # counter
    locus_tag_dict = {}  # recipient structure
    for feat in genome_seq_record.features:
        try:
            locustag = feat.qualifiers.get("locus_tag")[0]
            locus_tag_dict[loc_number] = (locustag, feat.type) 
            loc_number = loc_number + 1
        except TypeError:
            pass
    if len(locus_tag_dict) < 1:
            raise ValueError("no locus tags found!")
    print("filtering by feature of interest")
    filtered = {k: v[0] for k, v in locus_tag_dict.items() if v[1] == feature}
    if verbose: 
        for key in sorted(filtered):
                print("%s: %s" % (key, filtered[key]))
    return(filtered)


def stitch_together_target_regions(genome_sequence, coords, flanking=400, replace=True):
    """ 
    given a list from get_coords, usually of length 3 (16,5,and 23 rRNAs),
    return a string with the sequence of the region, replacing coding sequences
    with N's (or not, replace=False), and including the flanking regions 
    upstream and down.
    """
    region = ''
    for i in range(0, len(coords)):
        if i == 0:  # if first in list, start with flanking prefix
            start = coords[i][1][0] - flanking
            end = coords[i][1][0]
            region = "".join([region, str(genome_sequence[start:end])])
        if replace:
            start = coords[i][1][0]
            end = coords[i][1][1]
            region = "".join([region, str("N" * (end - start + 1))])
        else:
            region = "".join([region, str(genome_sequence[coords[i][1][0]: coords[i][1][1]])])               
        if i == (len(coords)-1):  # if last in list, tag on flanking suffix 
            start = coords[i][1][1]
            end = coords[i][1][1] + flanking
            region ="".join([region,  str(genome_sequence[start:end])])
        else:  # if not, get intergenic region, if it exists
            start = coords[i][1][1] + 1
            end = coords[i + 1][1][0] - 1
            region ="".join([region,  str(genome_sequence[start:end])])
    return(region)

def get_region_near_loc(genome_sequence, loc_list, feature_list, feature_regions,
                        feature_regs_len, within):
    """ The magic, part 2
    upstream, downstream, identity, identityN for reiong repped as N's
    TODO make this fail if it tries to reference outside of bounds
    """
    feature_region_options = ["upstream", "upstream_within", "downstream",
                              "downstream_within", "identity",
                              "identityN"]
    if not len(feature_list) == len(feature_regions) == len(feature_regs_len):
        raise ValueError("you supplied  as a feature list; length of feature_regions must \
                          have same length to match up properly")
    for i in range(0, len(feature_list)):
        # check feature
        if not feature_regions[i] in feature_region_options:
            raise ValueError(str(
                "{0} not in availible feature regions below:\n{1}".format(
                    feature_regions[i], feature_region_options)))
        for j in loc_list:
            if re.search(feature_list[i], str(j[3])):
                fiveprime = None
                threeprime = None
                dna = None
                if feature_regions[i] == "upstream" and j[2] == 1:
                    fiveprime = j[1][0] - feature_regs_len[i] - 1  # -1 accounts for 0 index
                    threeprime = j[1][0]
                    dna = genome_sequence[fiveprime:threeprime]
                    desc = "_from_%i_to_%i" % (fiveprime + 1, threeprime)
                elif feature_regions[i] == "upstream" and j[2] == -1:
                    fiveprime = j[1][1] + 1
                    threeprime = j[1][1] + feature_regs_len[i]
                    dna = genome_sequence[fiveprime:threeprime].reverse_complement().lower()
                    desc = "_from_%i_to_%i, revcomplemented" % (fiveprime + 1, threeprime)
                elif feature_regions[i] == "upstream_within" and j[2] == 1:
                    fiveprime = j[1][0] - feature_regs_len[i] - 1
                    threeprime = j[1][0] + within
                    dna = genome_sequence[fiveprime:threeprime]
                    desc = "_from_%i_to_%i" % (fiveprime + 1, threeprime)
                elif feature_regions[i] == "upstream_within" and j[2] == -1:
                    fiveprime = j[1][1] + 1 - within
                    threeprime = j[1][1] + feature_regs_len[i]
                    dna = genome_sequence[fiveprime:threeprime].reverse_complement().lower()
                    desc = "_from_%i_to_%i, revcomplemented" % (fiveprime + 1, threeprime)
                elif feature_regions[i] == "downstream" and j[2] == 1:
                    fiveprime = j[1][1] + 1
                    threeprime = j[1][1] + feature_regs_len[i]
                    dna = genome_sequence[fiveprime:threeprime]
                    desc = "_from_%i_to_%i" % (fiveprime + 1, threeprime)
                elif feature_regions[i] == "downstream" and j[2] == -1:
                    fiveprime = j[1][0] - feature_regs_len[i] - 1
                    threeprime = j[1][0]
                    dna = genome_sequence[fiveprime:threeprime].reverse_complement().lower()
                    desc = "_from_%i_to_%i, revcomplemented" % (fiveprime + 1, threeprime)
                elif feature_regions[i] == "downstream_within" and j[2] == 1:
                    fiveprime = j[1][1] + 1 - within
                    threeprime = j[1][1] + feature_regs_len[i]
                    dna = genome_sequence[fiveprime:threeprime]
                    desc = "_from_%i_to_%i" % (fiveprime + 1, threeprime)
                elif feature_regions[i] == "downstream_within" and j[2] == -1:
                    fiveprime = j[1][0] - feature_regs_len[i] - 1
                    threeprime = j[1][0] + within
                    dna = genome_sequence[fiveprime:threeprime].reverse_complement().lower()
                    desc = "_from_%i_to_%i, revcomplemented" % (fiveprime + 1, threeprime)
                elif feature_regions[i] == "identity" and j[2] == 1:
                    fiveprime = j[1][0] - 1
                    threeprime = j[1][1]
                    dna = genome_sequence[fiveprime:threeprime]
                    desc = "_from_%i_to_%i" % (fiveprime + 1, threeprime)
                elif feature_regions[i] == "identity" and j[2] == -1:
                    fiveprime = j[1][0] - 1
                    threeprime = j[1][1]
                    dna = genome_sequence[fiveprime:threeprime].reverse_complement().lower()
                    desc = "_from_%i_to_%i, revcomplemented" % (fiveprime + 1, threeprime)
                elif feature_regions[i] == "identityN":
                    fiveprime = j[1][0] - 1
                    threeprime = j[1][1]
                    dna = str("N" + ("N" * (fiveprime - threeprime)))
                    desc = "_from_%i_to_%i" % (fiveprime + 1, threeprime)
                else:
                    raise ValueError("Why isnt %s caught earlier" %
                                     feature_regions[i])
                j.append(dna)
                j.append(desc)
            else:
                continue
    rm_list = []  # remove entries without sequences
    for i in loc_list:
        try:
            len(i[5])
        except IndexError:
            rm_list.append(i)
    trimmed_list = [x for x in loc_list if x not in rm_list]
    return(trimmed_list)



#def merge_neighboring_sequences(coords_with_seq, window_width=500):
#    group_number = 10
#    coords_with_seq[0].append(10)
#    for i in range(1, len(coords_with_seq)):
#        if coords_with_seq[i][1][0] - coords_with_seq[i-1][1][1] < window_width:
#            coords_with_seq[i-1].append(group_number)
#            print(group_number)
#        else:
#            group_number = group_number + 10
#        ## take care of last one; below statement should always be true in a perfect world
#    if coords_with_seq[len(coords_with_seq)-1][1][0] - coords_with_seq[len(coords_with_seq)-2][1][1] < window_width :
#        coords_with_seq[len(coords_with_seq)-1].append(group_number)
#    for i in range(10, group_number+10, 10):
#        coords_gh = []
#        ncoords = 0
#        for j in coords_with_seq:
#            if j[-1] == i:
#                coords_gh.append(j[1])
#                ncoords = ncoords + 1
#        print(coords_gh)
#        print("hi")
#merge_neighboring_sequences(coords_with_seq, windxoow_width=500)
#%%

if __name__ == "__main__":
    raise("this doesnt work yet")
# make lists if colon:delimeted
    args = get_args(DEBUG=False)
    specific_features = args.specific_features.split(":")
    feature_regs = args.feature_regions.split(":")
    print(args)
    try:
        feature_regs_len = [int(x) for x in args.feature_regions_lengths.split(":")]
    except ValueError:
        print("feature_regions_lengths must be an integer, or integers separated" +
              "by colons, such as 400:33")
        sys.exit(1)
    if not len(specific_features) == len(feature_regs) == len(feature_regs_len):
        print("features, feature_refions, and feature_region_lengths must be equal")
        sys.exit(1)
    date = str(datetime.datetime.now().strftime('%Y%m%d'))
    if check_single_scaffold(args.genbank_genome):
        raise ValueError("You really should only give this genbank files with one \
                         scaffold for now. Problem with that? Fork off.")
    if not os.path.isdir(args.output):
        os.mkdir(args.output)
    genome_sequence = get_genbank_seq(args.genbank_genome)
    genome_record = get_genbank_record(args.genbank_genome)
    # TODO ofload this part til the #******# to  riboSelect.py
    lociDict = get_filtered_locus_tag_dict(genome_record)
    # for each specific feature, get number of occurances
    nfeatures_occur = []
    for i in specific_features:
        unique=[]
        for k,v in lociDict.items():
            t = v.find(i)
            unique.append(t)
        nfeatures_occur.append(len([x for x in unique if x!=-1]))
    print(str("number of occuraces of each specific feature: {0}; using " +
              " {1} clusters").format(nfeatures_occur, min(nfeatures_occur) ))
    pure_python_kmeans(lociDict.keys(), centers=min(nfeatures_occur))
    ## csv to dict
    with open('list.csv', mode='r') as infile:
        reader = csv.reader(infile)
        next(reader, None)  # skip the headers
        indexClusterDict = dict((rows[0],rows[1]) for rows in reader)
    ##
    print(lociDict)
    print(indexClusterDict)
    clusteredDict={}
    for k,v  in indexClusterDict.items():
        clusteredDict.setdefault(v,[]).append(lociDict[int(k)])
    print(clusteredDict)
    #******#
    print(clusteredDict["1"])
    regions = []
    for i in clusteredDict.keys():
        locus_tag_list = clusteredDict[i]
        coord_list = extract_coords_from_locus(genome_record,
                                               locus_tag_list=locus_tag_list,
                                               verbose=True)
        regions.append(stitch_together_target_regions(genome_sequence,
                                                     coords=coord_list,
                                                     flanking=400,
                                                     replace=True))
    output_index = 1
    for i in regions:
        filename = str("region_" + str(output_index))
        with open(os.path.join(args.output,
                               str(date + "_" + filename + "_riboSnag.fasta")),
                  "w") as outfile:
            SeqIO.write(SeqRecord(Seq(i,IUPAC.IUPACAmbiguousDNA()),
                                  id = str(str(output_index) + "_riboSnag"),
                                  description=""), outfile, "fasta")
            outfile.write('\n')
            output_index = output_index + 1
